from typing import Optional
import jsonschema


class MessageConverter:
    SCHEMA: Optional[dict] = None

    def validate(self, raw_data):
        if self.SCHEMA is None:
            raise NotImplementedError("please specify a SCHEMA")
        # FIXME: re-throw own exception type?
        jsonschema.validate(schema=self.SCHEMA, instance=raw_data)

    def to_python(self, raw_data):
        """
        validate and convert from JSONic data structures to python data structures
        """
        from libertem.io.dataset.base import IOBackend
        self.validate(raw_data)
        results = self.convert_to_python(raw_data)
        if "io_backend" in raw_data:
            results["io_backend"] = IOBackend.get_cls_by_id(raw_data["io_backend"])()
        return results

    def convert_to_python(self, raw_data):
        """
        override this method to provide your own conversion
        """
        return raw_data

    def from_python(self, raw_data):
        """
        convert from python data structures to JSONic data structures and validate
        """
        converted = self.convert_from_python(raw_data)
        self.validate(converted)
        return converted

    def convert_from_python(self, raw_data):
        """
        override this method to provide your own conversion
        """
        return raw_data



import os
import json
import hashlib
import sqlite3
import time
import glob
from typing import Optional, Union

import numpy as np

from .base import (
    DataSet, Partition, PartitionStructure
)
from libertem.io.dataset.cluster import ClusterDataSet
from sparseconverter import CUDA, NUMPY, ArrayBackend


class VerboseRow(sqlite3.Row):
    """sqlite3.Row with a __repr__"""
    def __repr__(self):
        return "<VerboseRow {!r}>".format(
            {
                k: self[k]
                for k in self.keys()
            },
        )


class CacheItem:
    """
    A CacheItem describes a single unit of data that is cached, in this case
    a partition of the CachedDataSet.
    """
    def __init__(self, dataset: str, partition: int, size: int, path: str):
        self.dataset = dataset  # dataset id string, for example the cache key
        self.partition = partition  # partition index as integer
        self.size = size  # partition size in bytes
        self.path = path  # full absolute path to the file for the partition
        self.is_orphan = False  # quack

    def __eq__(self, other):
        # dataset and partition are composite pk
        return self.dataset == other.dataset and self.partition == other.partition

    def __repr__(self):
        return "<CacheItem: %s/%d>" % (self.dataset, self.partition)

    @classmethod
    def from_row(cls, row):
        return cls(
            dataset=row["dataset"],
            partition=row["partition"],
            size=row["size"],
            path=row["path"]
        )


class OrphanItem:
    """
    An orphan, a file in the cache structure, which we don't know much about
    (only path and size)
    """
    def __init__(self, path, size):
        self.path = path
        self.size = size
        self.is_orphan = True

    def __eq__(self, other):
        return self.path == other.path

    def __repr__(self):
        return f"<OrphanItem: {self.path}>"

    @classmethod
    def from_row(cls, row):
        return cls(
            size=row["size"],
            path=row["path"]
        )


class CacheStats:
    def __init__(self, db_path):
        self._db_path = db_path
        self._conn = None
        self._connect()

    def _connect(self):
        conn = sqlite3.connect(self._db_path)
        conn.row_factory = VerboseRow
        self._conn = conn

    def close(self):
        self._conn.close()
        self._conn = None

    def initialize_schema(self):
        self._conn.execute("""
        CREATE TABLE IF NOT EXISTS stats (
            dataset VARCHAR NOT NULL,           -- dataset id, for example the cache key
            partition INTEGER NOT NULL,         -- partition index as integer
            hits INTEGER NOT NULL,              -- access counter
            size INTEGER NOT NULL,              -- in bytes
            last_access REAL NOT NULL,          -- float timestamp, like time.time()
            path VARCHAR NOT NULL,              -- full path to the file for this partition
            PRIMARY KEY (dataset, partition)
        );""")
        self._conn.execute("""
        CREATE TABLE IF NOT EXISTS orphans (
            path VARCHAR NOT NULL,              -- full path to the orphaned file
            size INTEGER NOT NULL,              -- in bytes
            PRIMARY KEY (path)
        );""")
        self._conn.execute("PRAGMA user_version = 1;")

    def _have_item(self, cache_item: CacheItem):
        rows = self._conn.execute("""
        SELECT hits FROM stats
        WHERE dataset = ? AND partition = ?
        """, [cache_item.dataset, cache_item.partition]).fetchall()
        return len(rows) > 0

    def record_hit(self, cache_item: CacheItem):
        now = time.time()
        self._conn.execute("BEGIN")
        if not self._have_item(cache_item):
            self._conn.execute("""
            INSERT INTO stats (partition, dataset, hits, size, last_access, path)
            VALUES (?, ?, 1, ?, ?, ?)
            """, [cache_item.partition, cache_item.dataset,
                  cache_item.size, now, cache_item.path])
        else:
            self._conn.execute("""
            UPDATE stats
            SET hits = MAX(hits + 1, 1), last_access = ?
            WHERE dataset = ? AND partition = ?
            """, [now, cache_item.dataset, cache_item.partition])
        self._conn.execute("DELETE FROM orphans WHERE path = ?", [cache_item.path])
        self._conn.commit()

    def record_miss(self, cache_item: CacheItem):
        now = time.time()

        self._conn.execute("BEGIN")
        if not self._have_item(cache_item):
            self._conn.execute("""
            INSERT INTO stats (partition, dataset, hits, size, last_access, path)
            VALUES (?, ?, 0, ?, ?, ?)
            """, [cache_item.partition, cache_item.dataset, cache_item.size,
                  now, cache_item.path])
        else:
            self._conn.execute("""
            UPDATE stats
            SET hits = 0, last_access = ?
            WHERE dataset = ? AND partition = ?
            """, [now, cache_item.dataset, cache_item.partition])
        self._conn.execute("DELETE FROM orphans WHERE path = ?", [cache_item.path])
        self._conn.commit()

    def record_eviction(self, cache_item: Union[CacheItem, OrphanItem]):
        if cache_item.is_orphan:
            self.remove_orphan(cache_item)
        else:
            self._conn.execute("""
            DELETE FROM stats
            WHERE partition = ? AND dataset = ?
            """, [cache_item.partition, cache_item.dataset])
            self._conn.commit()

    def maybe_orphan(self, orphan: OrphanItem):
        """
        Create an entry for a file we don't have any statistics about, after checking
        the stats table for the given path.
        Getting a conflict here means concurrently running maybe_orphan processes,
        so we can safely ignore it.
        """
        exists = len(self._conn.execute("""
        SELECT 1 FROM stats WHERE path = ?
        """, [orphan.path]).fetchall()) > 0
        if not exists:
            self._conn.execute("""
            INSERT OR IGNORE INTO orphans (path, size)
            VALUES (?, ?)
            """, [orphan.path, orphan.size])
            self._conn.commit()
            return orphan

    def get_orphans(self):
        cursor = self._conn.execute("SELECT path, size FROM orphans ORDER BY size DESC")
        return [
            OrphanItem.from_row(row)
            for row in cursor
        ]

    def remove_orphan(self, path: str):
        self._conn.execute("""
        DELETE FROM orphans WHERE path = ?
        """, [path])
        self._conn.commit()

    def get_stats_for_dataset(self, cache_key):
        """
        Return dataset cache stats as dict mapping partition ids to dicts of their
        properties (keys: size, last_access, hits)
        """
        cursor = self._conn.execute("""
        SELECT partition, path, size, hits, last_access
        FROM stats
        WHERE dataset = ?
        """, [cache_key])
        return {
            row["partition"]: self._format_row(row)
            for row in cursor.fetchall()
        }

    def query(self, sql, args=None):
        """
        Custom sqlite query, returns a sqlite3 Cursor object
        """
        if args is None:
            args = []
        return self._conn.execute(sql, args)

    def _format_row(self, row):
        return {
            "path": row["path"],
            "size": row["size"],
            "last_access": row["last_access"],
            "hits": row["hits"],
        }

    def get_used_capacity(self):
        size = self._conn.execute("""
        SELECT SUM(size) AS "total_size" FROM stats;
        """).fetchone()["total_size"]
        size_orphans = self._conn.execute("""
        SELECT SUM(size) AS "total_size" FROM orphans;
        """).fetchone()["total_size"]
        return (size or 0) + (size_orphans or 0)


class CacheStrategy:
    def get_victim_list(self, cache_key: str, size: int, stats: CacheStats):
        """
        Return a list of `CacheItem`s that should be deleted to make a new item
        with size in bytes `size`.
        """
        raise NotImplementedError()


class LRUCacheStrategy(CacheStrategy):
    def __init__(self, capacity: int):
        self._capacity = capacity
        super().__init__()

    def get_victim_list(self, cache_key: str, size: int, stats: CacheStats):
        """
        Return a list of `CacheItem`s that should be deleted to make
        place for `partition`.
        """
        # LRU with the following modifications:
        # 1) Don't evict from the same dataset, as our accesses
        #    are highly correlated in a single dataset
        # 2) Include orphaned files as preferred victims
        # 3) TODO: work in an estimated miss cost (challenge: estimate I/O cost
        #    independently from whatever calculation the user decides to run!)
        if self.sufficient_space_for(size, stats):
            return []
        victims = []
        space_to_free = size - self.get_available(stats)

        orphans = stats.get_orphans()

        candidates = stats.query("""
        SELECT dataset, partition, size, path
        FROM stats
        WHERE dataset != ?
        ORDER BY last_access ASC
        """, [cache_key])

        to_check = orphans + [
            CacheItem.from_row(row)
            for row in candidates
        ]

        for item in to_check:
            if space_to_free <= 0:
                break
            victims.append(item)
            space_to_free -= item.size
        if space_to_free > 0:
            raise RuntimeError(
                "not enough cache capacity for the requested operation"
            )  # FIXME: exception class
        return victims

    def sufficient_space_for(self, size: int, stats: CacheStats):
        return size <= self.get_available(stats)

    def get_available(self, stats: CacheStats):
        """
        available cache capacity in bytes
        """
        return self._capacity - self.get_used(stats)

    def get_used(self, stats: CacheStats):
        """
        used cache capacity in bytes
        """
        return stats.get_used_capacity()


class Cache:
    """
    Cache object, to be used on a worker node. The interface used by `Partition`\\ s
    to manage the cache. May directly remove files, directories, etc.
    """
    def __init__(self, stats: CacheStats, strategy: CacheStrategy):
        self._stats = stats
        self.strategy = strategy

    def record_hit(self, cache_item: CacheItem):
        self._stats.record_hit(cache_item)

    def record_miss(self, cache_item: CacheItem):
        self._stats.record_miss(cache_item)

    def evict(self, cache_key: str, size: int):
        """
        Make place for `size` bytes which will be used
        by the dataset identified by the `cache_key`.
        """
        victims = self.strategy.get_victim_list(cache_key, size, self._stats)
        for cache_item in victims:
            # if it has been deleted by the user, we don't care and just remove
            # the record from the database:
            if os.path.exists(cache_item.path):
                os.unlink(cache_item.path)
            self._stats.record_eviction(cache_item)

    def collect_orphans(self, base_path: str):
        """
        Check the filesystem structure and record all partitions
        that are missing in the db as orphans, to be deleted on demand.
        """
        # the structure here is: {base_path}/{dataset_cache_key}/parts/*
        orphans = []
        for path in glob.glob(os.path.join(glob.escape(base_path), "*", "parts", "*")):
            size = os.stat(path).st_size
            res = self._stats.maybe_orphan(OrphanItem(path=path, size=size))
            if res is not None:
                orphans.append(res)
        return orphans


class CachedDataSet(DataSet):
    """
    Cached DataSet.

    Assumes the source DataSet is significantly slower than the cache location
    (otherwise, it may result in memory pressure, as we don't use direct I/O
    to write to the cache.)

    Parameters
    ----------
    source_ds : DataSet
        DataSet on slower file system

    cache_path : str
        Where should the cache be written to? Should be a directory on a fast
        local device (i.e. NVMe SSD if possible, local hard drive if it is faster than network)
        A subdirectory will be created for each dataset.

    strategy : CacheStrategy
        A class implementing a cache eviction strategy, for example LRUCacheStrategy
    """
    def __init__(self, source_ds, cache_path, strategy, io_backend=None):
        super().__init__(io_backend=io_backend)
        self._source_ds = source_ds
        self._cache_path = cache_path
        self._cache_key = self._make_cache_key(source_ds.get_cache_key())
        self._path = os.path.join(cache_path, self._cache_key)
        self._cluster_ds = None
        self._executor = None
        self._cache_strategy = strategy

    def _make_cache_key(self, inp):
        inp_as_str = json.dumps(inp)
        return hashlib.sha256(inp_as_str.encode("utf-8")).hexdigest()

    def initialize(self, executor):
        source_structure = PartitionStructure.from_ds(self._source_ds)
        executor.run_each_host(self._ensure_cache_structure)
        cluster_ds = ClusterDataSet(
            path=self._path,
            structure=source_structure,
        )
        cluster_ds.check_valid()
        self._cluster_ds = cluster_ds.initialize(executor=executor)
        self._executor = executor
        return self

    def _get_db_path(self):
        return os.path.join(self._cache_path, "cache.db")

    def _ensure_cache_structure(self):
        os.makedirs(self._path, exist_ok=True)

        cache_stats = CacheStats(self._get_db_path())
        cache_stats.initialize_schema()
        cache = Cache(stats=cache_stats, strategy=self._cache_strategy)
        cache.collect_orphans(self._cache_path)

    @property
    def dtype(self):
        return self._cluster_ds.dtype

    @property
    def shape(self):
        return self._cluster_ds.shape

    @classmethod
    def get_msg_converter(cls):
        raise NotImplementedError(
            "not directly usable from web API"
        )

    def check_valid(self):
        # TODO: validate self._cache_path, what else?
        # - ask cache backend if things look valid (i.e. sidecar cache info is OK)
        return True

    def get_partitions(self):
        for idx, (source_part, cluster_part) in enumerate(zip(self._source_ds.get_partitions(),
                                                              self._cluster_ds.get_partitions())):
            yield CachedPartition(
                source_part=source_part,
                cluster_part=cluster_part,
                meta=cluster_part.meta,
                partition_slice=cluster_part.slice,
                cache_key=self._cache_key,
                cache_strategy=self._cache_strategy,
                db_path=self._get_db_path(),
                idx=idx,
                io_backend=self.get_io_backend(),
                decoder=None,
            )

    def evict(self, executor):
        for _ in executor.run_each_partition(self.get_partitions(),
                                             lambda p: p.evict(), all_nodes=True):
            pass

    def __repr__(self):
        return "<CachedDataSet dtype={} shape={} source_ds={} cache_path={} path={}>".format(
            self.dtype, self.shape, self._source_ds, self._cache_path, self._path
        )


class CachedPartition(Partition):
    def __init__(self, source_part, cluster_part, meta, partition_slice,
                 cache_key, cache_strategy, db_path, idx, io_backend, decoder):
        super().__init__(
            meta=meta,
            partition_slice=partition_slice,
            io_backend=io_backend,
            decoder=decoder
        )
        self._source_part = source_part
        self._cluster_part = cluster_part
        self._cache_key = cache_key
        self._cache_strategy = cache_strategy
        self._db_path = db_path
        self._idx = idx

    def _get_cache(self):
        cache_stats = CacheStats(self._db_path)
        return Cache(stats=cache_stats, strategy=self._cache_strategy)

    def _sizeof(self):
        return self.slice.shape.size * np.dtype(self.dtype).itemsize

    def _write_tiles_noroi(self, wh, source_tiles, dest_dtype):
        """
        Write tiles from source_tiles to the cache. After each tile is written, yield
        it for further processing, potentially doing dtype conversion on the fly.
        """
        with wh:
            miss_tiles = wh.write_tiles(source_tiles)
            if np.dtype(dest_dtype) != np.dtype(self._cluster_part.dtype):
                for tile in miss_tiles:
                    yield tile.astype(dest_dtype)
            else:
                yield from miss_tiles

    def _write_tiles_roi(self, wh, source_tiles, cached_tiles):
        """
        Get source tiles without roi, read and cache whole partition, then
        read all tiles selected via roi from the cache (_cluster_part aka cached_tiles).
        """
        with wh:
            for tile in wh.write_tiles(source_tiles):
                pass
        yield from cached_tiles

    def get_tiles(self, tiling_scheme, dest_dtype="float32", roi=None,
            array_backend: Optional[ArrayBackend] = None):
        if array_backend is None:
            array_backend = self.meta.array_backends[0]
        assert array_backend in (NUMPY, CUDA)
        cache = self._get_cache()
        cached_tiles = self._cluster_part.get_tiles(
            tiling_scheme=tiling_scheme,
            dest_dtype=dest_dtype,
            roi=roi,
        )

        cache_item = CacheItem(
            dataset=self._cache_key,
            partition=self._idx,
            path=self._cluster_part.get_canonical_path(),
            size=self._sizeof(),
        )
        if self._cluster_part._have_data():
            yield from cached_tiles
            cache.record_hit(cache_item)
        else:
            cache.evict(cache_key=self._cache_key, size=self._sizeof())
            # NOTE: source_tiles are in native dtype!
            source_tiles = self._source_part.get_tiles(
                tiling_scheme=tiling_scheme,
                dest_dtype=self._cluster_part.dtype,
                roi=None,  # NOTE: want to always cache the whole dataset, thus roi=None
            )
            wh = self._cluster_part.get_write_handle()
            if roi is None:
                yield from self._write_tiles_noroi(wh, source_tiles, dest_dtype)
            else:
                yield from self._write_tiles_roi(wh, source_tiles, cached_tiles)
            cache.record_miss(cache_item)

    def get_locations(self):
        """
        returns locations where this partition is cached
        """
        return self._cluster_part.get_locations()

    def evict(self):
        self._cluster_part.delete()



import functools
import warnings
import logging
from typing_extensions import TypedDict

import psutil

import numba.cuda


logger = logging.getLogger(__name__)


try:
    import cupy
except ModuleNotFoundError:
    cupy = None
except ImportError as e:
    # Cupy can be a bit fragile; allow running LiberTEM with
    # messed-up installation
    warnings.warn(repr(e), RuntimeWarning)
    cupy = None


class DetectResult(TypedDict):
    cpus: list[int]
    cudas: list[int]
    has_cupy: bool


def detect() -> DetectResult:
    '''
    Detect which devices are present

    .. versionadded:: 0.6.0

    Returns
    -------
    dict
        Dictionary with keys :code:`'cpus'` and :code:`'cudas'`
        Each containing
        a list of devices. Only physical CPU cores are counted, i.e. no
        hyperthreading.
        Additionally it has the key :code:`'has_cupy'`, which signals
        if cupy is installed and available.
    '''
    cores = psutil.cpu_count(logical=False)
    if cores is None:
        cores = 2
    try:
        cudas = [device.id for device in numba.cuda.gpus]
    except numba.cuda.CudaSupportError as e:
        # Continue running without GPU or in case of errors
        cudas = []
        logger.info(repr(e))
    return {
        "cpus": list(range(cores)),
        "cudas": cudas,
        "has_cupy": has_cupy(),
    }


@functools.cache
def has_cupy():
    '''
    Probe if :code:`cupy` was loaded successfully.

    .. versionadded:: 0.6.0

    CuPy is an optional dependency with special integration for UDFs. See
    :ref:`udf cuda` for details.
    '''
    if cupy is None:
        return False
    try:
        cupy.cuda
        cupy.array(cupy.zeros((1,)))
        return True
    except Exception as e:  # possibly: AttributeError or CompileException
        warnings.warn(repr(e), RuntimeWarning)
        return False



import numpy as np

import pytest

from libertem.analysis.clust import ClusterAnalysis
from libertem.io.dataset.memory import MemoryDataSet
from libertem.executor.base import AsyncAdapter


@pytest.mark.asyncio
@pytest.mark.slow
async def test_cluster_analysis(inline_executor):
    data = np.zeros([16, 16, 8, 8]).astype(np.float32)
    data[:, 2, 2] = 7
    # adding strong non-zero order diffraction peaks for 0:3 frames
    data[0:3, 0, 0] = 2
    data[0:3, 4, 4] = 2
    # adding weak non-zero order diffraction peaks for 0:3 frames
    data[3:6, 2, 0] = 1
    data[3:6, 2, 4] = 1

    dataset = MemoryDataSet(data=data, tileshape=(1, 8, 8),
                            num_partitions=1, sig_dims=2)

    executor = AsyncAdapter(wrapped=inline_executor)

    analysis = ClusterAnalysis(dataset=dataset, parameters={
        'n_peaks': 23,
        'n_clust': 7,
        'cy': 3,
        'cx': 3,
        'ri': 1,
        'ro': 5,
        'delta': 0.05,
        'min_dist': 1,
    })

    uuid = 'bd3b39fb-0b34-4a45-9955-339da6501bbb'

    async def send_results(results, finished):
        pass

    await analysis.controller(
        cancel_id=uuid, executor=executor,
        job_is_cancelled=lambda: False,
        send_results=send_results,
    )



import typing


NodeType = typing.Union[None, 'FileTree']


class FileTree:
    """
    Construct a FileTree node

    Parameters
    ----------

    low
        First frame contained in this file

    high
        First index of the next file

    value
        The corresponding file object

    idx
        The index of the file object in the fileset

    left
        Nodes with a lower low

    right
        Nodes with a higher low
    """
    def __init__(self, low: int, high: int, value: typing.Any, idx: int,
                 left: NodeType, right: NodeType):
        if low >= high:
            raise ValueError("low should be < high")
        self.low = low
        self.high = high
        self.value = value
        self.idx = idx
        self.left = left
        self.right = right

    @classmethod
    def make(cls, files):
        """
        build a balanced binary tree by bisecting the files list
        """

        def _make(files):
            if len(files) == 0:
                return None
            mid = len(files) // 2
            idx, value = files[mid]

            return FileTree(
                low=value.start_idx,
                high=value.end_idx,
                value=value,
                idx=idx,
                left=_make(files[:mid]),
                right=_make(files[mid + 1:]),
            )
        return _make(list(enumerate(files)))

    def search_start(self, value):
        """
        search a node that has start_idx <= value && end_idx > value
        """
        if self.low <= value and self.high > value:
            return self.idx, self.value
        elif self.low > value:
            return self.left.search_start(value)
        else:
            return self.right.search_start(value)

    def __str__(self):
        return self.to_string()

    def to_string(self, depth=0):
        padsingle = 4 * ' '
        pad = (depth * padsingle)
        str_left = (self.left is not None and self.left.to_string(depth + 1) or 'None')
        str_right = (self.right is not None and self.right.to_string(depth + 1) or 'None')
        str_self = f"(d={depth} l={self.low}, i={self.idx}, h={self.high})"
        return f"{pad}{str_left}\n{pad}{str_self}\n{pad}{str_right}"



import numpy as np
from libertem.masks import _make_circular_mask
from .raw import PickFrameAnalysis
from .helper import GeneratorHelper


class PickfftTemplate(GeneratorHelper):

    short_name = "pickfft"

    def __init__(self, params):
        self.params = params

    def get_dependency(self):
        return ["from libertem.analysis import PickFFTFrameAnalysis",
                "from matplotlib import colors"]

    # FIXME write and include documentation
    def get_docs(self):
        docs = ["# PICK FFT Analysis"]
        return '\n'.join(docs)

    def get_analysis(self):
        temp_analysis = [
                f"pickfft_analysis = PickFFTFrameAnalysis(dataset=ds, parameters={self.params})",
                "pickfft_result = ctx.run(pickfft_analysis, progress=True)"
                ]
        return '\n'.join(temp_analysis)

    def get_plot(self):
        plot = ["plt.figure()",
                "plt.imshow(pickfft_result.intensity, norm=colors.LogNorm())",
                "plt.colorbar()"]
        return ['\n'.join(plot)]


class PickFFTFrameAnalysis(PickFrameAnalysis, id_="PICK_FFT_FRAME"):
    def get_udf_results(self, udf_results, roi, damage):
        data = udf_results['intensity'].data[0]
        real_rad = self.parameters.get("real_rad")
        real_center = (self.parameters.get("real_centery"), self.parameters.get("real_centerx"))
        if data.dtype.kind == 'c':
            return self.get_generic_results(data, damage=damage)
        if not (real_center is None or real_rad is None):
            sigshape = data.shape
            real_mask = 1-1*_make_circular_mask(
                real_center[1], real_center[0], sigshape[1], sigshape[0], real_rad
            )
            fft_data = np.fft.fftshift(abs(np.fft.fft2(data*real_mask)))
        else:
            fft_data = np.fft.fftshift(abs(np.fft.fft2(data)))
        return self.get_generic_results(fft_data, damage=damage)

    @classmethod
    def get_template_helper(cls):
        return PickfftTemplate



import io
import os
import numpy as np
import nbformat
from temp_utils import _get_hdf5_params
from libertem.analysis import SumSigAnalysis
from libertem.web.notebook_generator.notebook_generator import notebook_generator
from nbconvert.preprocessors import ExecutePreprocessor
import pytest


@pytest.mark.slow
def test_sum_sig_default(hdf5_ds_2, tmpdir_factory, lt_ctx, local_cluster_url):
    datadir = tmpdir_factory.mktemp('template_tests')

    conn = {'connection': {'type': 'tcp', 'address': local_cluster_url}}
    path = hdf5_ds_2.path
    dataset = _get_hdf5_params(path)

    analysis = [{
            "analysisType": 'SUM_SIG',
            "parameters": {}
    }]

    notebook = notebook_generator(conn, dataset, analysis, save=True)
    notebook = io.StringIO(notebook.getvalue())
    nb = nbformat.read(notebook, as_version=4)
    ep = ExecutePreprocessor(timeout=600)
    ep.preprocess(nb, {"metadata": {"path": datadir}})
    data_path = os.path.join(datadir, 'sumsig_result.npy')
    results = np.load(data_path)

    analysis = SumSigAnalysis(
                          dataset=hdf5_ds_2,
                          parameters={}
                        )
    expected = lt_ctx.run(analysis)
    assert np.allclose(
        results,
        expected['intensity'].raw_data,
    )



import os
import json
import threading
import itertools
from unittest import mock
import sys
from typing import Optional

import cloudpickle
import numpy as np
from numpy.testing import assert_allclose
import pytest
import h5py

from libertem.io.dataset.hdf5 import H5DataSet
from libertem.analysis.sum import SumAnalysis
from libertem.udf.sumsigudf import SumSigUDF
from libertem.udf.auto import AutoUDF
from libertem.io.dataset.base import TilingScheme, DataSetException
from libertem.common import Shape
from libertem.common.math import flat_nonzero
from libertem.io.dataset.base import Negotiator
from libertem.udf import UDF

from utils import _naive_mask_apply, _mk_random, PixelsumUDF
from utils import dataset_correction_verification, roi_as_sparse


def test_hdf5_apply_masks_1(lt_ctx, hdf5_ds_1):
    mask = _mk_random(size=(16, 16))
    with hdf5_ds_1.get_reader().get_h5ds() as h5ds:
        data = h5ds[:]
        expected = _naive_mask_apply([mask], data)
    analysis = lt_ctx.create_mask_analysis(
        dataset=hdf5_ds_1, factories=[lambda: mask]
    )
    results = lt_ctx.run(analysis)

    assert np.allclose(
        results.mask_0.raw_data,
        expected
    )


def test_hdf5_3d_apply_masks(lt_ctx, hdf5_ds_3d):
    mask = _mk_random(size=(16, 16))
    with hdf5_ds_3d.get_reader().get_h5ds() as h5ds:
        data = h5ds[:]
        expected = _naive_mask_apply([mask], data.reshape((1, 17, 16, 16)))
    analysis = lt_ctx.create_mask_analysis(
        dataset=hdf5_ds_3d, factories=[lambda: mask]
    )
    results = lt_ctx.run(analysis)

    assert np.allclose(
        results.mask_0.raw_data,
        expected
    )


def test_hdf5_5d_apply_masks(lt_ctx, hdf5_ds_5d):
    mask = _mk_random(size=(16, 16))
    with hdf5_ds_5d.get_reader().get_h5ds() as h5ds:
        data = h5ds[:]
        expected = _naive_mask_apply([mask], data.reshape((1, 135, 16, 16))).reshape((3, 5, 9))
    analysis = lt_ctx.create_mask_analysis(
        dataset=hdf5_ds_5d, factories=[lambda: mask]
    )
    results = lt_ctx.run(analysis)

    print(results.mask_0.raw_data.shape, expected.shape)

    assert np.allclose(
        results.mask_0.raw_data,
        expected
    )


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
def test_correction(hdf5_ds_1, lt_ctx, with_roi):
    ds = hdf5_ds_1

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None

    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


def test_read_1(lt_ctx, hdf5):
    ds = H5DataSet(
        path=hdf5.filename, ds_path="data",
    )
    ds = ds.initialize(lt_ctx.executor)
    tileshape = Shape(
        (16,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )
    for p in ds.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            print(t.tile_slice)


def test_read_2(lt_ctx, hdf5):
    ds = H5DataSet(
        path=hdf5.filename, ds_path="data",
    )
    ds = ds.initialize(lt_ctx.executor)
    tileshape = Shape(
        (16,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )
    for p in ds.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            print(t.tile_slice)


def test_read_3(lt_ctx, random_hdf5):
    # try with smaller partitions:
    ds = H5DataSet(
        path=random_hdf5.filename, ds_path="data",
        target_size=4096
    )
    ds = ds.initialize(lt_ctx.executor)
    tileshape = Shape(
        (16,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )
    for p in ds.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            print(t.tile_slice)


def test_pickle_ds(lt_ctx, hdf5_ds_1):
    pickled = cloudpickle.dumps(hdf5_ds_1)
    loaded = cloudpickle.loads(pickled)

    assert loaded._dtype is not None

    # let's keep the pickled dataset size small-ish:
    assert len(pickled) < 1 * 1024


def test_cloudpickle(lt_ctx, hdf5):
    ds = H5DataSet(
        path=hdf5.filename, ds_path="data", target_size=512*1024*1024
    )

    pickled = cloudpickle.dumps(ds)
    loaded = cloudpickle.loads(pickled)

    assert loaded._dtype is None
    assert loaded._shape is None
    repr(loaded)

    ds = ds.initialize(lt_ctx.executor)

    pickled = cloudpickle.dumps(ds)
    loaded = cloudpickle.loads(pickled)

    assert loaded._dtype is not None
    assert loaded._shape is not None
    loaded.shape
    loaded.dtype
    repr(loaded)

    # let's keep the pickled dataset size small-ish:
    assert len(pickled) < 1 * 1024


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_roi_1(hdf5, lt_ctx, as_sparse):
    ds = H5DataSet(
        path=hdf5.filename, ds_path="data",
    )
    ds = ds.initialize(lt_ctx.executor)
    p = next(ds.get_partitions())
    roi = np.zeros(p.meta.shape.flatten_nav().nav, dtype=bool)
    roi[0] = 1
    tiles = []
    tileshape = Shape(
        (16,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )
    if as_sparse:
        roi = roi_as_sparse(roi)
    for tile in p.get_tiles(tiling_scheme=tiling_scheme, dest_dtype="float32", roi=roi):
        print("tile:", tile)
        tiles.append(tile)
    assert len(tiles) == 1
    assert tiles[0].tile_slice.shape.nav.size == 1
    assert tuple(tiles[0].tile_slice.shape.sig) == (16, 16)
    assert tiles[0].tile_slice.origin == (0, 0, 0)


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_roi_3(hdf5, lt_ctx, as_sparse):
    ds = H5DataSet(
        path=hdf5.filename, ds_path="data",
        target_size=12800*2,
    )
    ds = ds.initialize(lt_ctx.executor)
    roi = np.zeros(ds.shape.flatten_nav().nav, dtype=bool)
    roi[24] = 1

    tileshape = Shape(
        (16,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )

    tiles = []
    if as_sparse:
        roi = roi_as_sparse(roi)
    for p in ds.get_partitions():
        for tile in p.get_tiles(tiling_scheme=tiling_scheme, dest_dtype="float32", roi=roi):
            print("tile:", tile)
            tiles.append(tile)
    assert len(tiles) == 1
    assert tiles[0].tile_slice.shape.nav.size == 1
    assert tuple(tiles[0].tile_slice.shape.sig) == (16, 16)
    assert tiles[0].tile_slice.origin == (0, 0, 0)
    assert np.allclose(tiles[0].data, hdf5['data'][4, 4])


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_roi_4(hdf5, lt_ctx, as_sparse):
    ds = H5DataSet(
        path=hdf5.filename, ds_path="data",
        target_size=12800*2,
    )
    ds = ds.initialize(lt_ctx.executor)
    roi = np.random.choice(size=ds.shape.flatten_nav().nav, a=[True, False])

    sum_udf = lt_ctx.create_sum_analysis(dataset=ds)
    if as_sparse:
        roi = roi_as_sparse(roi)
    sumres = lt_ctx.run(sum_udf, roi=roi)['intensity']

    assert np.allclose(
        sumres,
        np.sum(hdf5['data'][:].reshape(25, 16, 16)[flat_nonzero(roi), ...], axis=0)
    )


def test_roi_5(hdf5, lt_ctx):
    ds = H5DataSet(
        path=hdf5.filename, ds_path="data",
        target_size=12800*2,
    )
    ds = ds.initialize(lt_ctx.executor)
    roi = np.random.choice(size=ds.shape.flatten_nav().nav, a=[True, False])

    udf = SumSigUDF()
    sumres = lt_ctx.run_udf(dataset=ds, udf=udf, roi=roi)['intensity']

    assert np.allclose(
        sumres.raw_data,
        np.sum(hdf5['data'][:][roi.reshape(5, 5), ...], axis=(1, 2))
    )


def test_pick(hdf5, lt_ctx):
    ds = H5DataSet(
        path=hdf5.filename, ds_path="data",
    )
    ds = ds.initialize(lt_ctx.executor)
    assert len(ds.shape) == 4
    print(ds.shape)
    pick = lt_ctx.create_pick_analysis(dataset=ds, x=2, y=3)
    lt_ctx.run(pick)


def test_diags(hdf5, lt_ctx):
    ds = H5DataSet(
        path=hdf5.filename, ds_path="data",
    )
    ds = ds.initialize(lt_ctx.executor)
    print(ds.diagnostics)


def test_check_valid(hdf5, lt_ctx):
    ds = H5DataSet(
        path=hdf5.filename, ds_path="data",
    )
    ds = ds.initialize(lt_ctx.executor)
    assert ds.check_valid()


def test_timeout_1(hdf5, lt_ctx):
    with mock.patch('h5py.File.visititems', side_effect=TimeoutError("too slow")):
        params = H5DataSet.detect_params(hdf5.filename, executor=lt_ctx.executor)["parameters"]
        assert list(params.keys()) == ["path"]

        ds = H5DataSet(
            path=hdf5.filename, ds_path="data",
        )
        ds = ds.initialize(lt_ctx.executor)
        diags = ds.diagnostics
        print(diags)


def test_timeout_2(hdf5, lt_ctx):
    print(threading.enumerate())
    with mock.patch('libertem.io.dataset.hdf5.current_time', side_effect=[1, 60]):
        params = H5DataSet.detect_params(hdf5.filename, executor=lt_ctx.executor)["parameters"]
        assert list(params.keys()) == ["path"]

    ds = H5DataSet(
        path=hdf5.filename, ds_path="data",
    )
    ds = ds.initialize(lt_ctx.executor)

    print(threading.enumerate())
    with mock.patch('libertem.io.dataset.hdf5.current_time', side_effect=[30, 90]):
        diags = ds.diagnostics
        print(diags)


@pytest.mark.parametrize("mnp", [None, 1, 4])
def test_roi_2(random_hdf5, lt_ctx, mnp):
    ds = H5DataSet(
        path=random_hdf5.filename, ds_path="data",
        min_num_partitions=mnp,
    )
    ds = ds.initialize(lt_ctx.executor)

    roi = {
        "shape": "disk",
        "cx": 2,
        "cy": 2,
        "r": 1,
    }
    analysis = SumAnalysis(dataset=ds, parameters={
        "roi": roi,
    })

    print(analysis.get_roi())

    results = lt_ctx.run(analysis)

    # let's draw a circle!
    mask = np.full((5, 5), False)
    mask[1, 2] = True
    mask[2, 1:4] = True
    mask[3, 2] = True

    print(mask)

    assert mask.shape == (5, 5)
    assert mask.dtype == bool

    reader = ds.get_reader()
    with reader.get_h5ds() as h5ds:
        data = np.array(h5ds)

        # applying the mask flattens the first two dimensions, so we
        # only sum over axis 0 here:
        expected = data[mask, ...].sum(axis=(0,))

        assert expected.shape == (16, 16)
        assert results.intensity.raw_data.shape == (16, 16)

        # is not equal to results without mask:
        assert not np.allclose(results.intensity.raw_data, data.sum(axis=(0, 1)))
        # ... but rather like `expected`:
        assert np.allclose(results.intensity.raw_data, expected)


def test_cache_key_json_serializable(hdf5, lt_ctx):
    ds = H5DataSet(
        path=hdf5.filename, ds_path="data",
    )
    ds = ds.initialize(lt_ctx.executor)
    json.dumps(ds.get_cache_key())


def test_no_tileshape(lt_ctx, hdf5_3d):
    ds = lt_ctx.load('HDF5', path=hdf5_3d.filename, ds_path='/data')

    udf = PixelsumUDF()
    lt_ctx.run_udf(udf=udf, dataset=ds)


def test_scheme_idx(lt_ctx, hdf5):
    ds = H5DataSet(
        path=hdf5.filename, ds_path="data",
    )
    ds = ds.initialize(lt_ctx.executor)
    p = next(ds.get_partitions())

    sig_shape = tuple(ds.shape.sig)
    tileshape = Shape(
        (16,) + sig_shape[:-1] + (sig_shape[-1] // 2,),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)

    for tile, expected_idx in zip(tiles, itertools.cycle([0, 1])):
        print(tile.scheme_idx, tile.tile_slice)
        assert tile.scheme_idx == expected_idx


def test_hdf5_with_2d_shape(lt_ctx, hdf5_2d):
    # regression test for error message:
    # sync_offset should be in (0, 0), which is (-image_count, image_count)
    with pytest.raises(DataSetException) as e:
        hdf5_ds_2d = lt_ctx.load("hdf5", path=hdf5_2d.filename, ds_path="data")

        # these lines are currently not reached yet, but will be if we decide to support 2D HDF5
        assert hdf5_ds_2d.shape.to_tuple() == (1, 16, 16)  # navigation shape is extended
        udf = PixelsumUDF()
        lt_ctx.run_udf(udf=udf, dataset=hdf5_ds_2d)

    assert e.match("2D HDF5 files are currently not supported")


@pytest.mark.parametrize('chunks', [
    (1, 3, 16, 16),
    (1, 6, 16, 16),
    (1, 4, 16, 16),
    (1, 16, 16, 16),
])
def test_chunked(lt_ctx, tmpdir_factory, chunks):
    datadir = tmpdir_factory.mktemp('data')
    filename = os.path.join(datadir, 'hdf5-test-chunked.h5')
    data = _mk_random((16, 16, 16, 16), dtype=np.float32)

    with h5py.File(filename, "w") as f:
        f.create_dataset("data", data=data, chunks=chunks)

    ds = lt_ctx.load("hdf5", path=filename)
    udf = PixelsumUDF()
    res = lt_ctx.run_udf(udf=udf, dataset=ds)['pixelsum']
    assert np.allclose(
        res,
        np.sum(data, axis=(2, 3))
    )


@pytest.fixture(scope='module')
def shared_random_data():
    return _mk_random(size=(16, 16, 256, 256), dtype='float32')


@pytest.mark.parametrize('udf', [
    SumSigUDF(),
    AutoUDF(f=lambda frame: frame.sum()),
])
@pytest.mark.parametrize('chunks', [
    (3, 3, 32, 32),
    (3, 6, 32, 32),
    (3, 4, 32, 32),
    (1, 4, 32, 32),
    (1, 16, 32, 32),

    (3, 3, 256, 256),
    (3, 6, 256, 256),
    (3, 4, 256, 256),
    (1, 4, 256, 256),
    (1, 16, 256, 256),

    (3, 3, 128, 256),
    (3, 6, 128, 256),
    (3, 4, 128, 256),
    (1, 4, 128, 256),
    (1, 16, 128, 256),

    (3, 3, 32, 128),
    (3, 6, 32, 128),
    (3, 4, 32, 128),
    (1, 4, 32, 128),
    (1, 16, 32, 128),
])
def test_chunked_weird(lt_ctx, tmpdir_factory, chunks, udf, shared_random_data):
    datadir = tmpdir_factory.mktemp('data')
    filename = os.path.join(datadir, 'weirdly-chunked-256-256.h5')
    data = shared_random_data

    with h5py.File(filename, "w") as f:
        f.create_dataset("data", data=data, chunks=chunks)

    ds = lt_ctx.load("hdf5", path=filename)

    base_shape = ds.get_base_shape(roi=None)
    print(base_shape)

    res = lt_ctx.run_udf(dataset=ds, udf=udf)
    assert len(res) == 1
    res = next(iter(res.values()))
    assert_allclose(
        res,
        np.sum(data, axis=(2, 3))
    )

    os.unlink(filename)


@pytest.mark.parametrize('in_dtype', [
    np.float32,
    np.float64,
    np.uint16,
])
@pytest.mark.parametrize('read_dtype', [
    np.float32,
    np.float64,
    np.uint16,
])
@pytest.mark.parametrize('use_roi', [
    True, False
])
def test_hdf5_result_dtype(lt_ctx, tmpdir_factory, in_dtype, read_dtype, use_roi):
    datadir = tmpdir_factory.mktemp('data')
    filename = os.path.join(datadir, 'result-dtype-checks.h5')
    data = _mk_random((2, 2, 4, 4), dtype=in_dtype)

    with h5py.File(filename, "w") as f:
        f.create_dataset("data", data=data)

    ds = lt_ctx.load("hdf5", path=filename)

    if use_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool).reshape((-1,))
        roi[0] = 1
    else:
        roi = None
    udfs = [SumSigUDF()]  # need to have at least one UDF
    p = next(ds.get_partitions())
    neg = Negotiator()
    tiling_scheme = neg.get_scheme(
        udfs=udfs,
        dataset=ds,
        approx_partition_shape=p.shape,
        read_dtype=read_dtype,
        roi=roi,
        corrections=None,
    )
    tile = next(p.get_tiles(tiling_scheme=tiling_scheme, roi=roi, dest_dtype=read_dtype))
    assert tile.dtype == np.dtype(read_dtype)


class UDFWithLargeDepth(UDF):
    def process_tile(self, tile):
        pass

    def get_tiling_preferences(self):
        return {
            "depth": 128,
            "total_size": UDF.TILE_SIZE_BEST_FIT,
        }


def test_hdf5_tileshape_negotation(lt_ctx, tmpdir_factory):
    # try to hit the third case in _get_subslices:
    datadir = tmpdir_factory.mktemp('data')
    filename = os.path.join(datadir, 'tileshape-neg-test.h5')
    data = _mk_random((4, 100, 256, 256), dtype=np.uint16)

    with h5py.File(filename, "w") as f:
        f.create_dataset("data", data=data, chunks=(2, 32, 32, 32))

    ds = lt_ctx.load("hdf5", path=filename)

    udfs = [UDFWithLargeDepth()]
    p = next(ds.get_partitions())
    neg = Negotiator()
    tiling_scheme = neg.get_scheme(
        udfs=udfs,
        dataset=ds,
        approx_partition_shape=p.shape,
        read_dtype=np.float32,
        roi=None,
        corrections=None,
    )
    assert len(tiling_scheme) > 1
    next(p.get_tiles(tiling_scheme=tiling_scheme, roi=None, dest_dtype=np.float32))


def test_scheme_too_large(hdf5_ds_1):
    partitions = hdf5_ds_1.get_partitions()
    p = next(partitions)
    depth = p.shape[0]

    # we make a tileshape that is too large for the partition here:
    tileshape = Shape(
        (depth + 1,) + tuple(hdf5_ds_1.shape.sig),
        sig_dims=hdf5_ds_1.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=hdf5_ds_1.shape,
    )

    # tile shape is clamped to partition shape.
    # in case of hdf5, it is even smaller than the
    # partition, as the depth from the negotiation
    # is overridden:
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert t.tile_slice.shape[0] <= hdf5_ds_1.shape[0]


def test_hdf5_macrotile(lt_ctx, tmpdir_factory):
    datadir = tmpdir_factory.mktemp('data')
    filename = os.path.join(datadir, 'macrotile-1.h5')
    data = _mk_random((128, 128, 4, 4), dtype=np.float32)

    with h5py.File(filename, "w") as f:
        f.create_dataset("data", data=data)

    ds = lt_ctx.load("hdf5", path=filename)
    ds.set_num_cores(4)

    partitions = ds.get_partitions()
    p0 = next(partitions)
    m0 = p0.get_macrotile()
    assert m0.tile_slice.origin == (0, 0, 0)
    assert m0.tile_slice.shape == p0.shape

    p1 = next(partitions)
    m1 = p1.get_macrotile()
    assert m1.tile_slice.origin == (p0.shape[0], 0, 0)
    assert m1.tile_slice.shape == p1.shape


def test_hdf5_macrotile_roi(lt_ctx, hdf5_ds_1):
    roi = np.random.choice(size=hdf5_ds_1.shape.flatten_nav().nav, a=[True, False])
    with hdf5_ds_1.get_reader().get_h5ds() as h5ds:
        data = h5ds[:]
        expected = data.reshape(hdf5_ds_1.shape.flatten_nav())[roi]
    partitions = hdf5_ds_1.get_partitions()
    p0 = next(partitions)
    m0 = p0.get_macrotile(roi=roi)
    assert_allclose(
        m0.data,
        expected
    )


def test_hdf5_macrotile_empty_roi(lt_ctx, hdf5_ds_1):
    roi = np.zeros(hdf5_ds_1.shape.flatten_nav().nav, dtype=bool)
    partitions = hdf5_ds_1.get_partitions()
    p0 = next(partitions)
    m0 = p0.get_macrotile(roi=roi)
    assert m0.shape == (0, 16, 16)
    assert_allclose(
        m0.data,
        0,
    )


def test_hdf5_filters(local_cluster_ctx, lt_ctx, tmpdir_factory):
    # Make sure it is either preloaded already
    # or not installed
    if 'hdf5plugin' not in sys.modules:
        with pytest.raises(ImportError):
            import hdf5plugin
    else:
        import hdf5plugin
        datadir = tmpdir_factory.mktemp('filtered')
        filename = os.path.join(datadir, 'filtered.h5')
        with h5py.File(filename, "w") as f:
            f.create_dataset(
                "data",
                data=np.ones((16, 16, 16, 16)),
                **hdf5plugin.LZ4()
            )
        for ctx in (local_cluster_ctx, lt_ctx):
            ds = ctx.load('HDF5', path=filename)
            res = ctx.run_udf(dataset=ds, udf=SumSigUDF())
            assert np.allclose(res['intensity'].raw_data, np.prod(ds.shape.sig))


def _place_results(
    data: np.ndarray,
    nav_shape: tuple[int],
    sync_offset: int = 0,
    roi: Optional[np.ndarray] = None
) -> np.ndarray:
    """
    Places ravelled data into an array of nav_shape
    following the sync_offset, roi and reshaping conventions
    of LiberTEM.

    roi.shape must equal nav_shape
    data.size can be smaller or larger than nav_shape.size
    the data will be clipped or filled with zeros appropriately
    """
    data = data.ravel()
    output_size = np.prod(nav_shape)
    output = np.zeros((output_size,), dtype=data.dtype)
    if sync_offset <= 0:
        nel = min(output[abs(sync_offset):].size, data.size)
        output[abs(sync_offset): abs(sync_offset) + nel] = data[:nel]
    else:
        nel = min(data[sync_offset:].size, output.size)
        output[:nel] = data[sync_offset:sync_offset + nel]
    output = output.reshape(nav_shape)
    if roi is not None:
        output[np.logical_not(roi)] = np.nan
    return output


@pytest.mark.parametrize(
    'file_nav_shape, nav_chunks, nav_shape,', [
        ((9,), (3,), (3, 3,)),
        ((4, 4), (2, 2), (2, 8)),
        ((4, 4, 2), (2, 4, 2), (32,)),
        ((16,), (4,), (3, 3,)),
        ((9, 9), (3, 9), (4, 4,)),
        ((9, 9), (3, 9), (16, 16)),
        ((125,), (10,), (5, 5)),
        ((125,), (10,), (100, 100)),
    ])
@pytest.mark.parametrize('roi', [
    None, True,
])
@pytest.mark.parametrize('sync_offset', [
    0, -5, 5,
])
def test_nav_reshape(lt_ctx, tmpdir_factory, file_nav_shape,
                     nav_chunks, nav_shape, roi, sync_offset):
    sig_shape = (16, 16)
    sig_chunks = (16, 16)
    sig_dims = len(sig_shape)
    frame_size_bytes = np.prod(sig_shape) * np.dtype(np.float32).itemsize
    # Choose a deliberately small partition size so that some
    # partition + sync_offset combinations mean some parts are unneeded
    target_size_bytes = 4 * frame_size_bytes

    datadir = tmpdir_factory.mktemp('data')
    file_shape = file_nav_shape + sig_shape
    filename = os.path.join(datadir, f'data_{"_".join(str(s) for s in file_shape)}.h5')
    data = np.random.uniform(size=file_shape).astype(np.float32)

    with h5py.File(filename, "w") as f:
        f.create_dataset("data", data=data, chunks=nav_chunks + sig_chunks)

    if roi:
        roi = np.random.choice([True, False], size=nav_shape).astype(bool)

    ds = lt_ctx.load(
        "hdf5",
        path=filename,
        nav_shape=nav_shape,
        sig_dims=sig_dims,
        target_size=target_size_bytes,
        sync_offset=sync_offset,
    )
    res = lt_ctx.run_udf(dataset=ds, udf=SumSigUDF(), roi=roi)

    full_sum_result = data.sum(axis=tuple(range(-1, -sig_dims - 1, -1))).ravel()
    sum_result = _place_results(
        full_sum_result,
        nav_shape,
        sync_offset=sync_offset,
        roi=roi,
    )

    assert np.allclose(
        res['intensity'].data,
        sum_result,
        equal_nan=roi is not None,
    )

    os.unlink(filename)


def test_sig_reshape_unsupported(lt_ctx, tmpdir_factory):
    file_shape = (8, 16, 16)
    sig_shape = (3, 3)
    sig_dims = 2

    datadir = tmpdir_factory.mktemp('data')
    filename = os.path.join(datadir, f'data_{"_".join(str(s) for s in file_shape)}.h5')
    data = np.random.uniform(size=file_shape).astype(np.float32)

    with h5py.File(filename, "w") as f:
        f.create_dataset("data", data=data)

    with pytest.raises(DataSetException):
        lt_ctx.load(
            "hdf5",
            path=filename,
            sig_shape=sig_shape,
            sig_dims=sig_dims,
        )

    os.unlink(filename)


def test_sync_offset_beyond_ds(lt_ctx, tmpdir_factory):
    file_shape = (8, 16, 16)
    sync_offset = 10

    datadir = tmpdir_factory.mktemp('data')
    filename = os.path.join(datadir, f'data_{"_".join(str(s) for s in file_shape)}.h5')
    data = np.random.uniform(size=file_shape).astype(np.float32)

    with h5py.File(filename, "w") as f:
        f.create_dataset("data", data=data)

    with pytest.raises(DataSetException):
        lt_ctx.load(
            "hdf5",
            path=filename,
            sync_offset=sync_offset,
        )

    os.unlink(filename)



from libertem.executor.utils import delayed_unpack


def unpack_example():
    nest = [5, 1, (2, 4), {'a': 7, 'b': 10}, 4, (3, 3), 1]
    flat = [5, 1, 2, 4, 7, 10, 4, 3, 3, 1]
    mapping = [[(list, 0)],
               [(list, 1)],
               [(list, 2), (tuple, 0)],
               [(list, 2), (tuple, 1)],
               [(list, 3), (dict, 'a')],
               [(list, 3), (dict, 'b')],
               [(list, 4)],
               [(list, 5), (tuple, 0)],
               [(list, 5), (tuple, 1)],
               [(list, 6)]]
    return nest, flat, mapping


def test_flatten_1():
    nest, flat, _ = unpack_example()
    flattened = delayed_unpack.flatten_nested(nest)
    assert len(flattened) == len(flat)
    assert all(a == b for a, b in zip(flat, flattened))


def test_mapping_1():
    nest, _, mapping = unpack_example()
    built_map = delayed_unpack.build_mapping(nest)
    assert len(built_map) == len(mapping)
    for built, true in zip(built_map, mapping):
        assert all(a == b for a, b in zip(built, true))


def test_rebuild_1():
    nest, flat, mapping = unpack_example()
    rebuilt = delayed_unpack.rebuild_nested(flat, mapping)
    assert len(rebuilt) == len(nest)
    for el_rebuilt, el_original in zip(rebuilt, nest):
        assert el_rebuilt == el_original


def test_empty_elements():
    ic = delayed_unpack.IgnoreClass
    nest = [(2, 4), {}, (4, []), 5]
    flat = [2, 4, ic, 4, ic, 5]
    mapping = [[(list, 0), (tuple, 0)],
               [(list, 0), (tuple, 1)],
               [(list, 1), (dict, ic)],
               [(list, 2), (tuple, 0)],
               [(list, 2), (tuple, 1), (list, ic)],
               [(list, 3)]]
    flattened = delayed_unpack.flatten_nested(nest)
    assert len(flattened) == len(flat)
    assert all(a == b for a, b in zip(flat, flattened))

    built_map = delayed_unpack.build_mapping(nest)
    assert len(built_map) == len(mapping)
    for built, true in zip(built_map, mapping):
        assert all(a == b for a, b in zip(built, true))

    rebuilt = delayed_unpack.rebuild_nested(flat, mapping)
    assert len(rebuilt) == len(nest)
    for el_rebuilt, el_original in zip(rebuilt, nest):
        assert el_rebuilt == el_original


def test_not_unpackable():
    nest = 5
    flat = [5]
    mapping = [None]

    flattened = delayed_unpack.flatten_nested(nest)
    assert len(flattened) == len(flat)
    assert all(a == b for a, b in zip(flat, flattened))

    built_map = delayed_unpack.build_mapping(nest)
    assert built_map == mapping

    rebuilt = delayed_unpack.rebuild_nested(flat, mapping)
    assert rebuilt == nest



import json

import pytest
import websockets

from utils import assert_msg
from aio_utils import (
    call_ca_rpc, create_connection, consume_task_results, create_default_dataset, create_analysis,
    create_update_compound_analysis, create_job_for_analysis, update_analysis,
)


pytestmark = [pytest.mark.web_api]


@pytest.mark.asyncio
async def test_run_job_1_sum(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(
        base_url, http_client, scheduler_url=local_cluster_url, token=default_token
    )

    print("checkpoint 1")

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, token=default_token,
        )

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, token=default_token,
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)

        # we are done with this job, clean up:
        async with http_client.delete(job_url) as resp:
            assert resp.status == 200
            assert_msg(await resp.json(), 'CANCEL_JOB')

        # also get rid of the dataset:
        async with http_client.delete(ds_url) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DELETE_DATASET')


@pytest.mark.asyncio
async def test_run_job_delete_ds(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    """
    main difference to test above: we just close the dataset without
    removing the job first. this tests another code path in `remove_dataset`
    """
    await create_connection(
        base_url, http_client, scheduler_url=local_cluster_url, token=default_token,
    )
    print("checkpoint 1")

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, token=default_token,
        )

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, token=default_token,
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)

        # get rid of the dataset without cancelling the job:
        async with http_client.delete(ds_url) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DELETE_DATASET')


@pytest.mark.asyncio
async def test_cancel_unknown_job(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(
        base_url, http_client, scheduler_url=local_cluster_url, token=default_token,
    )

    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        job_uuid = "un-kn-ow-n"
        job_url = f"{base_url}/api/jobs/{job_uuid}/?token={default_token}"

        # try to cancel unknown job:
        async with http_client.delete(job_url) as resp:
            assert resp.status == 200
            assert_msg(await resp.json(), 'CANCEL_JOB_FAILED', status='error')


@pytest.mark.asyncio
async def test_run_with_all_zeros_roi(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(
        base_url, http_client, scheduler_url=local_cluster_url, token=default_token,
    )
    print("checkpoint 1")

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, token=default_token,
        )

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, details={
                "analysisType": "SUM_FRAMES",
                "parameters": {
                    "roi": {
                        "shape": "disk",
                        "r": 0,
                        # setting the center outside of the frame
                        # means our ROI will be all-zero
                        "cx": -1,
                        "cy": -1,
                    }
                }
            }, token=default_token,
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)

        # we are done with this job, clean up:
        async with http_client.delete(job_url) as resp:
            assert resp.status == 200
            assert_msg(await resp.json(), 'CANCEL_JOB')

        # also get rid of the dataset:
        async with http_client.delete(ds_url) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DELETE_DATASET')


@pytest.mark.asyncio
async def test_run_job_update_analysis_parameters(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(
        base_url, http_client, scheduler_url=local_cluster_url, token=default_token,
    )
    print("checkpoint 1")

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, token=default_token,
        )

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, token=default_token,
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)

        analysis_data_updated = {
            "dataset": ds_uuid,
            "details": {
                "analysisType": "SUM_FRAMES",
                "parameters": {
                    "roi": {
                        "shape": "disk",
                        "r": 1,
                        "cx": 1,
                        "cy": 1,
                    }
                }
            }
        }

        async with http_client.put(analysis_url, json=analysis_data_updated) as resp:
            print(await resp.text())
            assert resp.status == 200
            resp_json = await resp.json()
            assert resp_json['status'] == "ok"

        msg = json.loads(await ws.recv())
        assert_msg(msg, 'ANALYSIS_UPDATED')
        assert msg['analysis'] == analysis_uuid
        assert msg['details']['parameters'] == {
            "roi": {
                "shape": "disk",
                "r": 1,
                "cx": 1,
                "cy": 1,
            },
        }

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)

        # we are done with this job, clean up:
        async with http_client.delete(job_url) as resp:
            assert resp.status == 200
            assert_msg(await resp.json(), 'CANCEL_JOB')

        # also get rid of the dataset:
        async with http_client.delete(ds_url) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DELETE_DATASET')


@pytest.mark.asyncio
async def test_analysis_removal(
    default_raw, base_url, http_client, server_port, shared_state, local_cluster_url, default_token
):
    await create_connection(
        base_url, http_client, scheduler_url=local_cluster_url, token=default_token,
    )

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        # compound analysis is first created without any analyses:
        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details=None, token=default_token,
        )

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, token=default_token,
        )

        # compound analysis is updated with the newly created analysis:
        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details={
                "mainType": "APPLY_RING_MASK",
                "analyses": [analysis_uuid]
            }, ca_uuid=ca_uuid, token=default_token,
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)

    # we are done with this analysis, clean up:
    # this will also remove the associated jobs
    async with http_client.delete(analysis_url) as resp:
        assert resp.status == 200
        assert_msg(await resp.json(), 'ANALYSIS_REMOVED')

    assert job_uuid not in shared_state.job_state.jobs
    assert analysis_uuid not in shared_state.analysis_state.analyses

    # also remove dataset:
    async with http_client.delete(ds_url) as resp:
        assert resp.status == 200
        assert_msg(await resp.json(), 'DELETE_DATASET')

    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')
        assert initial_msg == {
            "status": "ok",
            "messageType": "INITIAL_STATE",
            "datasets": [],
            "jobs": [],
            "analyses": [],
            "compoundAnalyses": [],
        }


@pytest.mark.asyncio
async def test_create_compound_analysis(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(
        base_url, http_client, scheduler_url=local_cluster_url, token=default_token,
    )

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        # compound analysis is first created without any analyses:
        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details=None, token=default_token,
        )

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, token=default_token,
        )

        # compound analysis is updated with the newly created analysis:
        _, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details={
                "mainType": "APPLY_RING_MASK",
                "analyses": [analysis_uuid]
            }, ca_uuid=ca_uuid, token=default_token,
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)


@pytest.mark.asyncio
async def test_rerun_com_analysis(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(
        base_url, http_client, scheduler_url=local_cluster_url, token=default_token,
    )

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        # compound analysis is first created without any analyses:
        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details=None, token=default_token,
        )

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, token=default_token,
            details={
                "analysisType": "CENTER_OF_MASS",
                "parameters": {
                    "shape": "com",
                    "cx": 128,
                    "cy": 128,
                    "r": 64,
                    "flip_y": False,
                    "scan_rotation": 0,
                }
            },
        )

        # compound analysis is updated with the newly created analysis:
        _, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details={
                "mainType": "CENTER_OF_MASS",
                "analyses": [analysis_uuid]
            }, ca_uuid=ca_uuid, token=default_token,
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)

        # update analysis and run again, this should cover the short circuit case
        await update_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, token=default_token,
            analysis_uuid=analysis_uuid,
            details={
                "analysisType": "CENTER_OF_MASS",
                "parameters": {
                    "shape": "com",
                    "cx": 128,
                    "cy": 128,
                    "r": 64,
                    "flip_y": False,
                    "scan_rotation": 88,
                }
            },
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)

        # and once again with changed parameters, this should cover the case
        # of having old results, but still needing a re-run:
        await update_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, token=default_token,
            analysis_uuid=analysis_uuid,
            details={
                "analysisType": "CENTER_OF_MASS",
                "parameters": {
                    "shape": "com",
                    "cx": 128,
                    "cy": 128,
                    "r": 48,
                    "flip_y": False,
                    "scan_rotation": 88,
                }
            },
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)


@pytest.mark.asyncio
async def test_com_guess_rpc(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(
        base_url, http_client, scheduler_url=local_cluster_url, token=default_token,
    )

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        # compound analysis is first created without any analyses:
        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details=None, token=default_token,
        )

        print("checkpoint 2.1")

        result = await call_ca_rpc(
            http_client, base_url, ca_uuid, "guess_parameters", token=default_token
        )
        assert result['status'] == "error"  # we don't have an analysis yet!

        print("checkpoint 2.5")

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, token=default_token,
            details={
                "analysisType": "CENTER_OF_MASS",
                "parameters": {
                    "shape": "com",
                    "cx": 128,
                    "cy": 128,
                    "r": 64,
                    "flip_y": False,
                    "scan_rotation": 0,
                }
            },
        )

        print("checkpoint 3")

        result = await call_ca_rpc(
            http_client, base_url, ca_uuid, "guess_parameters", token=default_token
        )
        assert result['status'] == "error"  # analysis is not added to the CA

        # compound analysis is updated with the newly created analysis:
        _, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details={
                "mainType": "CENTER_OF_MASS",
                "analyses": [analysis_uuid]
            }, ca_uuid=ca_uuid, token=default_token,
        )

        print("checkpoint 4")

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        print("checkpoint 4.5")

        await consume_task_results(ws, job_uuid)

        print("checkpoint 5")

        result = await call_ca_rpc(
            http_client, base_url, ca_uuid, "guess_parameters", token=default_token
        )
        # we don't currently care about what the results are, just that they are here:
        # (unit tests for the guessing take care of that detail)
        assert result['status'] == "ok"
        assert "guess" in result
        assert "cx" in result["guess"]
        assert "cy" in result["guess"]
        assert "scan_rotation" in result["guess"]
        assert "flip_y" in result["guess"]


@pytest.mark.asyncio
async def test_com_guess_rpc_need_to_run_job(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(
        base_url, http_client, scheduler_url=local_cluster_url, token=default_token,
    )

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        # compound analysis is first created without any analyses:
        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details=None, token=default_token,
        )

        print("checkpoint 2.1")

        result = await call_ca_rpc(
            http_client, base_url, ca_uuid, "guess_parameters", token=default_token
        )
        assert result['status'] == "error"  # we don't have an analysis yet!

        print("checkpoint 2.5")

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, token=default_token,
            details={
                "analysisType": "CENTER_OF_MASS",
                "parameters": {
                    "shape": "com",
                    "cx": 128,
                    "cy": 128,
                    "r": 64,
                    "flip_y": False,
                    "scan_rotation": 0,
                }
            },
        )

        print("checkpoint 3")

        result = await call_ca_rpc(
            http_client, base_url, ca_uuid, "guess_parameters", token=default_token
        )
        assert result['status'] == "error"  # analysis is not added to the CA

        # compound analysis is updated with the newly created analysis:
        _, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details={
                "mainType": "CENTER_OF_MASS",
                "analyses": [analysis_uuid]
            }, ca_uuid=ca_uuid, token=default_token,
        )

        print("checkpoint 4")

        result = await call_ca_rpc(
            http_client, base_url, ca_uuid, "guess_parameters", token=default_token
        )
        # we don't currently care about what the results are, just that they are here:
        # (unit tests for the guessing take care of that detail)
        assert result['status'] == "ok"
        assert "guess" in result
        assert "cx" in result["guess"]
        assert "cy" in result["guess"]
        assert "scan_rotation" in result["guess"]
        assert "flip_y" in result["guess"]



import io
import json

import h5py
import pytest
import websockets
import numpy as np
import nbformat

from utils import assert_msg

from aio_utils import (
    create_connection, consume_task_results, create_default_dataset, create_analysis,
    create_job_for_analysis, create_update_compound_analysis,
)
from libertem.io.writers.results import formats  # NOQA: F401
from libertem.io.writers.results.base import ResultFormatRegistry


pytestmark = [pytest.mark.web_api]


@pytest.mark.asyncio
async def test_download_hdf5(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(base_url, http_client, local_cluster_url, token=default_token)

    print("checkpoint 1")

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        await ws.ensure_open()
        print(ws.state)
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        print("checkpoint 3")

        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, token=default_token,
        )

        print("checkpoint 4")

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, token=default_token,
        )

        print("checkpoint 5")

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)

        download_url = "{}/api/compoundAnalyses/{}/analyses/{}/download/HDF5/?token={}".format(
            base_url, ca_uuid, analysis_uuid, default_token,
        )
        async with http_client.get(download_url) as resp:
            assert resp.status == 200
            raw_data = await resp.read()
            bio = io.BytesIO(raw_data)
            with h5py.File(bio, "r") as f:
                assert 'intensity' in f.keys()
                data = np.array(f['intensity'])
                default_raw_data = np.memmap(
                    filename=default_raw._path,
                    dtype=default_raw.dtype,
                    mode='r',
                    shape=tuple(default_raw.shape),
                )
                assert np.allclose(
                    default_raw_data.sum(axis=(0, 1)),
                    data
                )

        # we are done with this analysis, clean up:
        # this will also remove the associated jobs
        async with http_client.delete(analysis_url) as resp:
            assert resp.status == 200
            assert_msg(await resp.json(), 'ANALYSIS_REMOVED')

        # also get rid of the dataset:
        async with http_client.delete(ds_url) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DELETE_DATASET')


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'filetype',
    list(ResultFormatRegistry.get_available_formats().keys())
)
async def test_download_other_formats(
    default_raw, base_url, http_client, server_port, filetype, local_cluster_url, default_token,
):
    """
    This test just triggers download with different formats,
    it does not yet check the result files!
    """
    await create_connection(base_url, http_client, local_cluster_url, default_token)

    print("checkpoint 1")

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, token=default_token,
        )

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, token=default_token,
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)

        download_url = "{}/api/compoundAnalyses/{}/analyses/{}/download/{}/?token={}".format(
            base_url, ca_uuid, analysis_uuid, filetype, default_token,
        )
        async with http_client.get(download_url) as resp:
            assert resp.status == 200
            raw_data = await resp.read()
            bio = io.BytesIO(raw_data)
            bio.seek(0)
            assert len(bio.read()) > 0

        # we are done with this analysis, clean up:
        # this will also remove the associated jobs
        async with http_client.delete(analysis_url) as resp:
            assert resp.status == 200
            assert_msg(await resp.json(), 'ANALYSIS_REMOVED')

        # also get rid of the dataset:
        async with http_client.delete(ds_url) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DELETE_DATASET')

# TODO: add more checks that actually load the resulting files
# TODO: test multi-channel results (most important: HDF5 and TIFF)


@pytest.mark.asyncio
async def test_download_com(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(base_url, http_client, local_cluster_url, default_token)

    print("checkpoint 1")

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details={
                "mainType": "CENTER_OF_MASS",
                "analyses": [],
            }, token=default_token,
        )

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, details={
                "analysisType": "CENTER_OF_MASS",
                "parameters": {
                    "cx": 8,
                    "cy": 8,
                    "r": 32,
                }
            }, token=default_token,
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)

        download_url = "{}/api/compoundAnalyses/{}/analyses/{}/download/TIFF/?token={}".format(
            base_url, ca_uuid, analysis_uuid, default_token,
        )
        async with http_client.get(download_url) as resp:
            assert resp.status == 200

        # we are done with this analysis, clean up:
        # this will also remove the associated jobs
        async with http_client.delete(analysis_url) as resp:
            assert resp.status == 200
            assert_msg(await resp.json(), 'ANALYSIS_REMOVED')

        # also get rid of the dataset:
        async with http_client.delete(ds_url) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DELETE_DATASET')


@pytest.mark.asyncio
async def test_download_notebook(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(base_url, http_client, local_cluster_url, default_token)

    print("checkpoint 1")

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        print("checkpoint 2")
        initial_msg = json.loads(await ws.recv())
        assert_msg(initial_msg, 'INITIAL_STATE')

        ds_uuid, ds_url = await create_default_dataset(
            default_raw, ws, http_client, base_url, token=default_token,
        )

        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details={
                "mainType": "CENTER_OF_MASS",
                "analyses": [],
            }, token=default_token,
        )

        analysis_uuid, analysis_url = await create_analysis(
            ws, http_client, base_url, ds_uuid, ca_uuid, details={
                "analysisType": "CENTER_OF_MASS",
                "parameters": {
                    "cx": 8,
                    "cy": 8,
                    "r": 32,
                }
            }, token=default_token,
        )

        ca_uuid, ca_url = await create_update_compound_analysis(
            ws, http_client, base_url, ds_uuid, details={
                "mainType": "CENTER_OF_MASS",
                "analyses": [analysis_uuid],
            }, token=default_token,
        )

        job_uuid, job_url = await create_job_for_analysis(
            ws, http_client, base_url, analysis_uuid, token=default_token,
        )

        await consume_task_results(ws, job_uuid)

        download_url = "{}/api/compoundAnalyses/{}/download/notebook/?token={}".format(
            base_url, ca_uuid, default_token,
        )
        async with http_client.get(download_url) as resp:
            assert resp.status == 200
            raw_data = await resp.read()
            notebook = io.BytesIO(raw_data)
            notebook = nbformat.reads(notebook.getvalue(), as_version=4)
            assert nbformat.validate(notebook) is None

        # we are done with this analysis, clean up:
        # this will also remove the associated jobs
        async with http_client.delete(analysis_url) as resp:
            assert resp.status == 200
            assert_msg(await resp.json(), 'ANALYSIS_REMOVED')

        # also get rid of the dataset:
        async with http_client.delete(ds_url) as resp:
            assert resp.status == 200
            resp_json = await resp.json()
            assert_msg(resp_json, 'DELETE_DATASET')



import numpy as np

from libertem.analysis.rawfft import PickFFTFrameAnalysis
from libertem.io.dataset.memory import MemoryDataSet
from libertem.masks import _make_circular_mask

from utils import _mk_random


def test_pick_fft_defaults(lt_ctx):
    data = _mk_random([3*3, 8, 8], dtype=np.float32)
    dataset = MemoryDataSet(data=data, tileshape=(1, 8, 8),
                            num_partitions=2, sig_dims=2)
    analysis = PickFFTFrameAnalysis(dataset=dataset, parameters={
        'x': 1,
    })
    res = lt_ctx.run(analysis)
    fft_data = np.fft.fftshift(abs(np.fft.fft2(data[1])))

    assert np.allclose(res.intensity.raw_data, fft_data)


def test_pick_fft_masked(lt_ctx):
    data = _mk_random([3*3, 8, 8], dtype=np.float32)
    dataset = MemoryDataSet(data=data, tileshape=(1, 8, 8),
                            num_partitions=2, sig_dims=2)
    analysis = PickFFTFrameAnalysis(dataset=dataset, parameters={
        'x': 1,
        'real_rad': 1,
        'real_centerx': 1,
        'real_centery': 1,
    })
    real_mask = np.invert(_make_circular_mask(
        centerX=1, centerY=1, imageSizeX=8, imageSizeY=8, radius=1
    ))
    fft_data = np.fft.fftshift(abs(np.fft.fft2(data[1]*real_mask)))
    res = lt_ctx.run(analysis)

    assert np.allclose(res.intensity.raw_data, fft_data)



import sys

import numpy as np
import numba


@numba.njit(inline='always', cache=True)
def byteswap_2_straight(inp, out):
    for i in range(inp.shape[0] // 2):
        out[i * 2 + 0] = inp[i * 2 + 1]
        out[i * 2 + 1] = inp[i * 2 + 0]


@numba.njit(inline='always', cache=True)
def byteswap_2_decode(inp, out):
    for i in range(inp.shape[0] // 2):
        o0 = inp[i * 2 + 0] << 8
        o1 = inp[i * 2 + 1] << 0
        out[i] = o0 | o1


@numba.njit(inline='always', cache=True)
def byteswap_4_straight(inp, out):
    for i in range(inp.shape[0] // 4):
        out[i * 4 + 3] = inp[i * 4 + 0]
        out[i * 4 + 2] = inp[i * 4 + 1]
        out[i * 4 + 1] = inp[i * 4 + 2]
        out[i * 4 + 0] = inp[i * 4 + 3]


@numba.njit(inline='always', cache=True)
def byteswap_4_decode(inp, out):
    for i in range(inp.shape[0] // 4):
        o0 = inp[i * 4 + 0] << 24
        o1 = inp[i * 4 + 1] << 16
        o2 = inp[i * 4 + 2] << 8
        o3 = inp[i * 4 + 3] << 0
        out[i] = o0 + o1 + o2 + o3


@numba.njit(inline='always', cache=True)
def byteswap_8_straight(inp, out):
    for i in range(inp.shape[0] // 8):
        out[i * 8 + 7] = inp[i * 8 + 0]
        out[i * 8 + 6] = inp[i * 8 + 1]
        out[i * 8 + 5] = inp[i * 8 + 2]
        out[i * 8 + 4] = inp[i * 8 + 3]
        out[i * 8 + 3] = inp[i * 8 + 4]
        out[i * 8 + 2] = inp[i * 8 + 5]
        out[i * 8 + 1] = inp[i * 8 + 6]
        out[i * 8 + 0] = inp[i * 8 + 7]


@numba.njit(inline='always', cache=True)
def byteswap_8_decode(inp, out):
    for i in range(inp.shape[0] // 8):
        o0 = inp[i * 8 + 0] << 56
        o1 = inp[i * 8 + 1] << 48
        o2 = inp[i * 8 + 2] << 40
        o3 = inp[i * 8 + 3] << 32
        o4 = inp[i * 8 + 4] << 24
        o5 = inp[i * 8 + 5] << 16
        o6 = inp[i * 8 + 6] << 8
        o7 = inp[i * 8 + 7] << 0
        out[i] = o0 + o1 + o2 + o3 + o4 + o5 + o6 + o7


@numba.njit(inline='always', cache=True)
def default_decode(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    out[idx, :] = inp.view(native_dtype)


@numba.njit(inline='always', cache=True)
def decode_swap_2(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    byteswap_2_decode(inp, out=out[idx])


@numba.njit(inline='always', cache=True)
def decode_swap_4(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    byteswap_4_decode(inp, out=out[idx])


@numba.njit(inline='always', cache=True)
def decode_swap_8(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    byteswap_8_decode(inp, out=out[idx])


@numba.njit(inline='always', cache=True)
def decode_swap_only_2(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    byteswap_2_straight(inp, out=out[idx].view(np.uint8))


@numba.njit(inline='always', cache=True)
def decode_swap_only_4(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    byteswap_4_straight(inp, out=out[idx].view(np.uint8))


@numba.njit(inline='always', cache=True)
def decode_swap_only_8(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    byteswap_8_straight(inp, out=out[idx].view(np.uint8))


def _convert_byteorder_eq(order):
    if order != '=':
        return order
    return {
        'little': '<',
        'big': '>',
    }[sys.byteorder]


class Decoder:
    def do_clear(self):
        return False

    def get_native_dtype(self, inp_native_dtype, read_dtype):
        return inp_native_dtype

    def get_decode(self, native_dtype, read_dtype):
        raise NotImplementedError()


class DtypeConversionDecoder(Decoder):
    def _need_byteswap(self, native_dtype, read_dtype):
        native_dtype = np.dtype(native_dtype)
        read_dtype = np.dtype(read_dtype)
        nd_order = _convert_byteorder_eq(native_dtype.byteorder)
        rd_order = _convert_byteorder_eq(read_dtype.byteorder)
        return nd_order != rd_order and native_dtype.itemsize > 1

    def _swapping_decode(self, native_dtype):
        return {
            2: decode_swap_2,
            4: decode_swap_4,
            8: decode_swap_8,
        }[native_dtype.itemsize]

    def _swap_only_decode(self, native_dtype):
        return {
            2: decode_swap_only_2,
            4: decode_swap_only_4,
            8: decode_swap_only_8,
        }[native_dtype.itemsize]

    def get_decode(self, native_dtype, read_dtype):
        native_dtype = np.dtype(native_dtype)
        read_dtype = np.dtype(read_dtype)
        if not self._need_byteswap(native_dtype, read_dtype):
            return default_decode

        if native_dtype.kind in ('f', 'c'):
            # TODO: can implement f32->f32 and f64->f64 by straight swapping, and
            # other conversions via a two-step decoding process
            raise NotImplementedError(
                "byte swapping for floats not implemented yet"
            )

        return self._swapping_decode(native_dtype)

    def get_native_dtype(self, inp_native_dtype, read_dtype):
        if self._need_byteswap(inp_native_dtype, read_dtype):
            return np.dtype(np.uint8)
        return np.dtype(inp_native_dtype)



import logging
import inspect
from functools import partial

import numpy as np
import sparse

from libertem import masks
from libertem.common.math import prod
from .base import AnalysisResult, AnalysisResultSet
from .masks import BaseMasksAnalysis
from .helper import GeneratorHelper

log = logging.getLogger(__name__)


class RadialTemplate(GeneratorHelper):

    short_name = "radial"
    api = "create_radial_fourier_analysis"
    temp = GeneratorHelper.temp_analysis
    temp_analysis = temp + ["print(radial_result)"]

    def __init__(self, params):
        self.params = params

    def get_dependency(self):
        return [
            "import matplotlib.cm as cm",
            "from empyre.vis.colors import ColormapCubehelix, ColormapPerception"
        ]

    def get_docs(self):
        title = "Radial Fourier Analysis"
        from libertem.api import Context
        docs_rst = inspect.getdoc(Context.create_radial_fourier_analysis)
        docs = self.format_docs(title, docs_rst)
        return docs

    def convert_params(self):
        params = ['dataset=ds']
        for k in ['cx', 'cy', 'ri', 'ro', 'n_bins', 'max_order']:
            params.append(f'{k}={self.params[k]}')
        return ', '.join(params)

    def get_plot(self):
        cells = []
        cells.append([
            "fig, axes = plt.subplots()",
            'axes.set_title("dominant_0")',
            "plt.imshow(radial_result.dominant_0, cmap=cm.tab20, vmin=0, vmax=20)",
            "fig, axes = plt.subplots()",
            'axes.set_title("absolute_0_0")',
            "axes.imshow(radial_result.absolute_0_0)",
        ])
        cells.append([
            "imag = radial_result.complex_0_1.raw_data.imag",
            "real = radial_result.complex_0_1.raw_data.real",
            "ch = ColormapCubehelix(start=1, rot=1, minLight=0.5, maxLight=0.5, sat=2)",
            "fig, axes = plt.subplots()",
            'axes.set_title("complex_0_1")',
            "plt.imshow(ch.rgb_from_vector((real, imag, 0)))",
            "fig, axes = plt.subplots()",
            'axes.set_title("phase_0_1")',
            'plt.imshow(radial_result.phase_0_1.raw_data, cmap=ColormapPerception())'
        ])
        return ['\n'.join(cell) for cell in cells]

    def get_save(self):
        save = []
        channels = ["absolute_0_0", "absolute_0_1"]
        for channel in channels:
            result = f"radial_result['{channel}'].raw_data"
            save.append(f"np.save('radial_result_{channel}.npy', {result})")
        return '\n'.join(save)


class RadialFourierResultSet(AnalysisResultSet):
    """
    Result set of a :class:`RadialFourierAnalysis`

    Running a :class:`RadialFourierAnalysis` via :meth:`libertem.api.Context.run` on a dataset
    returns an instance of this class. It contains the Fourier coefficients
    for each bin. See :meth:`libertem.api.Context.create_radial_fourier_analysis` for
    available parameters and :ref:`radialfourier app` for a description of the application!

    .. versionadded:: 0.3.0

    Attributes
    ----------
    dominant_0, absolute_0_0, absolute_0_1, ..., absolute_0_<max_order>,\
    phase_0_0, ..., phase_0_<max_order>,\
    complex_0_0, ..., complex_0_<max_order>;\
    dominant_1, absolute_1_0, ..., complex_1_<max_order>;\
    dominant_<nbins-1>, ..., complex_<nbins-1>_<max_order> : libertem.analysis.base.AnalysisResult
        Results for each bin: dominant Fourier coefficient (indicates symmetry),
        absolute values of each Fourier coefficient,
        phase values of each Fourier coefficient, complex values of each Fourier coefficient.
        The results have the shape of the navigation dimension.
    raw_results : numpy.ndarray
        Complex numbers, shape (<n_bins>, <max_order + 1>, \\*(<ds.shape.nav>))
    """
    pass


def radial_mask_factory(detector_y, detector_x, cx, cy, ri, ro, n_bins, max_order, use_sparse):
    def stack():
        rings = masks.radial_bins(
            centerX=cx,
            centerY=cy,
            imageSizeX=detector_x,
            imageSizeY=detector_y,
            radius=ro,
            radius_inner=ri,
            n_bins=n_bins,
            use_sparse=use_sparse,
            dtype=np.complex64
        )

        orders = np.arange(max_order + 1)

        r, phi = masks.polar_map(
            centerX=cx,
            centerY=cy,
            imageSizeX=detector_x,
            imageSizeY=detector_y
        )
        modulator = np.exp(phi * orders[:, np.newaxis, np.newaxis] * 1j)

        if use_sparse:
            rings = rings.reshape((rings.shape[0], 1, *rings.shape[1:]))
            ring_stack = [rings] * len(orders)
            ring_stack = sparse.concatenate(ring_stack, axis=1)
            ring_stack *= modulator
        else:
            ring_stack = rings[:, np.newaxis, ...] * modulator
        return ring_stack.reshape((-1, detector_y, detector_x))
    return stack


class RadialFourierAnalysis(BaseMasksAnalysis, id_="RADIAL_FOURIER"):
    '''
    The Radial Fourier Analysis can be used to characterize
    atomic ordering in materials, in particular for low intensities where
    Fluctualtion EM :cite:`Gibson1997` has a hard time to distinguish speckle
    from shot noise. Reference :cite:`6980942` describes a previous application
    of this method to characterize features in images.

    This analysis doesn't use fast Fourier transforms, but calculates the
    Fourier coefficients using sparse matrices in a dot product following the
    `definition of Fourier series
    <https://en.wikipedia.org/wiki/Fourier_series#Complex-valued_functions>`_.

    See :meth:`libertem.api.Context.create_radial_fourier_analysis` for
    available parameters and :ref:`radialfourier app` for a description of the
    application!
    '''

    TYPE = 'UDF'

    def get_udf_results(self, udf_results, roi, damage):
        '''
        The AnalysisResults are calculated lazily in this function to reduce
        overhead.
        '''
        shape = tuple(self.dataset.shape.nav)
        # NOTE: transposed for historical reasons
        udf_results = udf_results['intensity'].data.reshape((prod(shape), -1)).T
        orders = self.parameters['max_order'] + 1
        n_bins = self.parameters['n_bins']
        udf_results = udf_results.reshape((n_bins, orders, *shape))

        def resultlist():
            from libertem.viz import CMAP_CIRCULAR_DEFAULT, visualize_simple, cmaps
            import matplotlib.cm as cm
            sets = []
            absolute = np.absolute(udf_results)
            normal = np.maximum(1, absolute[:, 0])
            # New local variable since this is a closure over damage
            dam = damage & np.all(np.isfinite(absolute), axis=(0, 1))
            normalized = absolute[:, 1:, ...] / normal[:, np.newaxis, ...]
            min_absolute = np.min(normalized[..., dam])
            max_absolute = np.max(normalized[..., dam])
            angle = np.angle(udf_results)
            threshold_map = absolute[:, 1:, ...].reshape((n_bins, -1)).max(axis=1) * 0.2
            below_threshold = np.all(
                absolute[:, 1:, ...] < threshold_map[:, np.newaxis, np.newaxis, np.newaxis],
                axis=1
            )
            dominant = np.argmax(absolute[:, 1:], axis=1) + 1
            dominant[below_threshold] = 0
            for b in range(n_bins):
                sets.append(
                    AnalysisResult(
                        raw_data=dominant[b],
                        visualized=partial(
                            visualize_simple, dominant[b], colormap=cm.tab20, vmin=0, vmax=20
                        ),
                        key="dominant_%s" % b,
                        title="dominant order of bin %s" % b,
                        desc="Dominant Fourier component",
                    )
                )
                sets.append(
                    AnalysisResult(
                        raw_data=absolute[b, 0],
                        visualized=partial(visualize_simple, absolute[b, 0], damage=dam),
                        key=f"absolute_{b}_{0}",
                        title=f"absolute of bin {b} order {0}",
                        desc="Absolute value of Fourier component",
                    )
                )
                for o in range(1, orders):
                    sets.append(
                        AnalysisResult(
                            raw_data=absolute[b, o],
                            visualized=partial(visualize_simple,
                                absolute[b, o] / normal[b], vmin=min_absolute, vmax=max_absolute,
                                damage=dam
                            ),
                            key=f"absolute_{b}_{o}",
                            title=f"absolute of bin {b} order {o}",
                            desc="Absolute value of Fourier component",
                        )
                    )
            for b in range(n_bins):
                for o in range(orders):
                    sets.append(
                        AnalysisResult(
                            raw_data=angle[b, o],
                            visualized=partial(visualize_simple,
                                angle[b, o], colormap=cmaps['perception_circular'],
                                damage=dam
                            ),
                            key=f"phase_{b}_{o}",
                            title=f"phase of bin {b} order {o}",
                            desc="Phase of Fourier component",
                        )
                    )
            for b in range(n_bins):
                data = udf_results[b, 0]
                f = partial(
                    CMAP_CIRCULAR_DEFAULT.rgb_from_vector,
                    (data.real, data.imag, 0),
                    vmax=np.max(np.abs(data[..., dam]))
                )
                sets.append(
                    AnalysisResult(
                        raw_data=data,
                        visualized=f,
                        key=f"complex_{b}_{0}",
                        title=f"bin {b} order {0}",
                        desc="Fourier component",
                    )
                )
                for o in range(1, orders):
                    data = udf_results[b, o] / normal[b]
                    f = partial(
                        CMAP_CIRCULAR_DEFAULT.rgb_from_vector,
                        (data.real, data.imag, 0), vmax=max_absolute
                    )
                    sets.append(
                        AnalysisResult(
                            raw_data=data,
                            visualized=f,
                            key=f"complex_{b}_{o}",
                            title=f"bin {b} order {o}",
                            desc="Fourier component",
                        )
                    )
            return sets
        return RadialFourierResultSet(resultlist, raw_results=udf_results)

    def get_mask_factories(self):
        if self.dataset.shape.sig.dims != 2:
            raise ValueError("can only handle 2D signals currently")

        (detector_y, detector_x) = self.dataset.shape.sig
        p = self.parameters

        return radial_mask_factory(
            detector_y=detector_y,
            detector_x=detector_x,
            cx=p['cx'],
            cy=p['cy'],
            ri=p['ri'],
            ro=p['ro'],
            n_bins=p['n_bins'],
            max_order=p['max_order'],
            use_sparse=p['use_sparse'],
        )

    def get_parameters(self, parameters):
        (detector_y, detector_x) = self.dataset.shape.sig

        cx = parameters.get('cx', detector_x / 2)
        cy = parameters.get('cy', detector_y / 2)
        ri = parameters.get('ri', 0)
        ro = parameters.get(
            'ro',
            masks.bounding_radius(cx, cy, detector_x, detector_y)
        )
        n_bins = parameters.get('n_bins', 1)
        max_order = parameters.get('max_order', 24)

        mask_count = n_bins * (max_order + 1)
        bin_width = (ro - ri) / n_bins
        bin_area = np.pi * ro**2 - np.pi * (ro - bin_width)**2
        stack_size = mask_count * detector_y * detector_x * 8

        default = 'scipy.sparse'
        # If the mask stack comfortably fits the L3 cache
        # FIXME more testing for optimum backend
        if stack_size < 2**18:
            default = False
        # Masks are actually dense
        elif bin_area / (detector_x * detector_y) > 0.05 and n_bins < 10:
            default = False

        use_sparse = parameters.get('use_sparse', default)
        return {
            'cx': cx,
            'cy': cy,
            'ri': ri,
            'ro': ro,
            'n_bins': n_bins,
            'max_order': max_order,
            'use_sparse': use_sparse,
            'mask_count': mask_count,
            'mask_dtype': np.complex64,
        }

    @classmethod
    def get_template_helper(cls):
        return RadialTemplate



# This module pre-loads modules that can have a long load time.
# To be used to avoid https://github.com/LiberTEM/LiberTEM/issues/218
# where loading modules makes workers unresponsive for too long.
# Usage: dask-worker [...] --preload libertem.preload [...]

# Disable flake8 because we import a lot of modules without using them.

# flake8: noqa

import numpy
import sparse
from matplotlib import colors, cm

try:
    import torch
except ImportError:
    pass

import libertem

try:
    import hdf5plugin
except ImportError:
    pass

import numba.cuda  # see also: #1432

from libertem.common.backend import set_file_limit
set_file_limit()



import time
import pytest
import typing
from contextlib import contextmanager

import libertem.api as lt
from libertem.udf.sum import SumUDF
from libertem.common.progress import TQDMProgressReporter, ProgressState
from libertem.common.executor import TaskCommHandler, WorkerQueue, WorkerQueueEmpty
from libertem.io.dataset.memory import MemoryDataSet

import utils

if typing.TYPE_CHECKING:
    from libertem.io.dataset.base.tiling import DataTile


class TrackingTQDM(TQDMProgressReporter):
    def __init__(self):
        super().__init__()
        self._history: list[ProgressState] = []

    def start(self, state):
        self._history.append(state)
        super().start(state)

    def update(self, state):
        self._history.append(state)
        super().update(state)

    def end(self, state):
        self._history.append(state)
        super().end(state)

    def print(self):
        for hist in self._history:
            print(hist)


class InlineProgressTaskCommHandler(TaskCommHandler):
    @contextmanager
    def monitor(self, queue: WorkerQueue):
        """
        Normally this would monitor the queue in the background
        and set up callbacks for progress, but instead just save
        a reference to the queue and wait, i.e. all messages received
        will build up in the queue and not be consumed
        """
        self._progress_queue = queue
        yield


class MemoryDataSetMockComms(MemoryDataSet):
    def get_task_comm_handler(self):
        """
        Save reference to patched comm handler so we can
        access the filled queue after the call to run_udf
        """
        self._comms_handler = InlineProgressTaskCommHandler()
        return self._comms_handler


def drain_queue(queue):
    while True:
        try:
            with queue.get(block=False) as ((topic, msg), _):
                yield topic, msg
        except WorkerQueueEmpty:
            break


class FastTime:
    """
    Time getter which increases by at least
    :code:`increment` seconds each call
    """
    def __init__(self, increment: float = 1.5):
        self._now = 0.
        self._increment = increment

    def __call__(self):
        self._now += self._increment
        return self._now


class WaitEndSumUDF(SumUDF):
    def postprocess(self):
        # Need time to let the message queue be processed
        # before the partition completes, as completion updates
        # happen synchronously on the main node and over-ride
        # further tile-level updates
        time.sleep(0.5)


def test_progress_inline_fasttime(lt_ctx, monkeypatch):
    """
    Tests that a comms failure will still correctly report
    Task completion, and that each partition sends signals
    when it begins and for each tile by patching the get_time
    function to increment quickly
    """
    fast_time = FastTime()
    monkeypatch.setattr('libertem.common.progress.get_time', fast_time)

    data = utils._mk_random(size=(4, 4, 16, 16), dtype='float32')
    ds = MemoryDataSetMockComms(
        data=data,
        num_partitions=2,
        tileshape=(4, 4, 16),
    )
    ds.initialize(lt_ctx)

    reporter = TrackingTQDM()
    udf = SumUDF()
    lt_ctx.run_udf(ds, udf, progress=reporter)

    assert reporter._bar.n == reporter._bar.total

    # These are the states which are generated synchronously
    # through a call to finalize_task() on the main thread
    # This tests the fallback mechanism in case comms fail
    states = reporter._history
    progress_id = states[0].progress_id
    start_progress = ProgressState(0., 16, 0, 0, 2, progress_id)
    assert start_progress in states
    part0_end_progress = ProgressState(8., 16, 1, 0, 2, progress_id)
    assert part0_end_progress in states
    part1_end_progress = ProgressState(16., 16, 2, 0, 2, progress_id)
    assert part1_end_progress in states

    # Get the queue to read the intermediate messages
    queue = ds._comms_handler._progress_queue
    # We expect 7 tile messages per partition as the first is skipped
    expected_order = ['partition_start'] + ['tile_complete'] * 7
    # Two partitions in this dataset
    expected_order = expected_order * 2

    # Drain the queue of messages and check the order
    current_part = None
    for (topic, message), expected_topic in zip(drain_queue(queue), expected_order):
        assert topic == expected_topic
        if topic == 'partition_start':
            current_part = message['ident']
        assert message['ident'] == current_part


class GoSlowSumUDF(WaitEndSumUDF):
    def process_tile(self, tile: 'DataTile'):
        super().process_tile(tile)
        # Sleep on the second tile
        # processed each partition
        try:
            if self._tiles_seen == 1:
                time.sleep(1.5)
        except AttributeError:
            self._tiles_seen = 0
        self._tiles_seen += 1


def _test_progress_slowudf(context: lt.Context):
    data = utils._mk_random(size=(4, 4, 16, 16), dtype='float32')
    ds = context.load(
        'memory',
        data=data,
        num_partitions=2,
        tileshape=(4, 4, 16),
    )

    reporter = TrackingTQDM()
    udf = GoSlowSumUDF()
    context.run_udf(ds, udf, progress=reporter)

    assert reporter._bar.n == reporter._bar.total
    num_part = ds.get_num_partitions()
    # Run start / stop + (part start/stop * num_part)
    min_num_messages = 2 + num_part * 2
    # We should recieve at least one extra message per part with this UDF
    # The only risk is if the postprocess time is not long enough for the
    # tile message to be processed!
    assert len(reporter._history) >= (min_num_messages + num_part)


@pytest.mark.slow
def test_progress_inline_slowudf(lt_ctx):
    _test_progress_slowudf(lt_ctx)


@pytest.mark.slow
def test_progress_concurrent_slowudf(concurrent_ctx):
    _test_progress_slowudf(concurrent_ctx)


@pytest.mark.slow
def test_progress_dask(dask_executor, default_raw):
    reporter = TrackingTQDM()
    udf = GoSlowSumUDF()
    runner_cls = dask_executor.get_udf_runner()
    runner = runner_cls([udf], progress_reporter=reporter)
    runner.run_for_dataset(
        default_raw,
        dask_executor,
        roi=None,
        progress=True
    )

    assert reporter._bar.n == reporter._bar.total
    num_part = default_raw.get_num_partitions()
    # Run start / stop + (part start / stop * num_part)
    min_num_messages = 2 + num_part * 2
    # default_raw doesn't have enough tiles to send intermediate
    # tile messages (depending on the system resources), but we can at least
    # check the comms are working by verifying we had min_num_messages
    assert len(reporter._history) >= min_num_messages


@pytest.mark.slow
def test_progress_pipelined(default_raw):
    with lt.Context.make_with("pipelined") as ctx:
        reporter = TrackingTQDM()
        udf = GoSlowSumUDF()
        ctx.run_udf(default_raw, udf, progress=reporter)

        assert reporter._bar.n == reporter._bar.total
        num_part = default_raw.get_num_partitions()
        # Run start / stop + (part start / stop * num_part)
        min_num_messages = 2 + num_part * 2
        # default_raw doesn't have enough tiles to send intermediate
        # tile messages (depending on the system resources), but we can at least
        # check the comms are working by verifying we had min_num_messages
        assert len(reporter._history) >= min_num_messages



import h5py
import numpy as np
from libertem.io.dataset.hdf5 import H5DataSet
from libertem.executor.inline import InlineJobExecutor


def _get_hdf5_params(path):
    return {
        "type": "HDF5",
        "params": {
            "path": path,
            "ds_path": "/data"
            },
    }


def create_random_hdf5(path):
    with h5py.File(path, 'w') as f:
        sample_data = np.random.randn(16, 16, 16, 16).astype("float32")
        f.create_dataset("data", (16, 16, 16, 16), data=sample_data)
        # read and provide the ds
    ds = H5DataSet(path=path, ds_path='data')
    ds = ds.initialize(InlineJobExecutor())
    return ds



from io import BytesIO
from typing import Optional

from PIL import Image


def encode_image(result, save_kwargs: Optional[dict] = None) -> BytesIO:
    """
    Save the RGBA data in ``result`` to an image with parameters ``save_kwargs``
    passed to ``PIL.Image.save``.

    Parameters
    ----------
    result : numpy.ndarray
        Array of RGB values; shape (height, width, 3)

    save_kwargs : dict or None
        dict of kwargs passed to Pillow when saving the image, can be used to set
        the file format, quality, ...

    Returns
    -------

    BytesIO
        a buffer containing the result image (as PNG/JPG/... depending on save_kwargs)
    """
    if save_kwargs is None:
        save_kwargs = {'format': 'png'}
    # see also: https://stackoverflow.com/a/10967471/540644
    im = Image.fromarray(result)
    buf = BytesIO()
    im = im.convert(mode="RGB")
    im.save(buf, **save_kwargs)
    buf.seek(0)
    return buf



import os
import sys
import json
import random

import numpy as np
import pytest
import cloudpickle

from libertem.udf.sum import SumUDF
from libertem.udf.raw import PickUDF

from libertem.analysis.raw import PickFrameAnalysis
from libertem.io.dataset.raw import RAWDatasetParams, RawFileDataSet
from libertem.io.dataset.base import (
    TilingScheme, BufferedBackend, MMapBackend, DirectBackend,
)
from libertem.common import Shape
from libertem.common.buffers import reshaped_view
from libertem.udf.sumsigudf import SumSigUDF

from utils import dataset_correction_verification, ValidationUDF, roi_as_sparse


@pytest.fixture
def raw_dataset_8x8x8x8(lt_ctx, raw_data_8x8x8x8_path):
    ds = RawFileDataSet(
        path=raw_data_8x8x8x8_path,
        nav_shape=(8, 8),
        sig_shape=(8, 8),
        dtype="float32",
    )
    ds.set_num_cores(4)
    ds = ds.initialize(lt_ctx.executor)

    return ds


def test_simple_open(default_raw):
    assert tuple(default_raw.shape) == (16, 16, 128, 128)


@pytest.mark.skipif(sys.platform.startswith("darwin"), reason="No general sparse support on OS X")
def test_large_pick(large_raw, lt_ctx):
    y, x = large_raw.shape.nav
    dy, dx = large_raw.shape.sig
    analysis = lt_ctx.create_pick_analysis(large_raw, y=y-1, x=x-1)
    result = lt_ctx.run(analysis)
    assert result.intensity.raw_data.shape == tuple(large_raw.shape.sig)


def test_check_valid(default_raw):
    default_raw.check_valid()


@pytest.mark.with_numba
def test_read(default_raw):
    partitions = default_raw.get_partitions()
    p = next(partitions)
    # FIXME: partition shape can vary by number of cores
    # assert tuple(p.shape) == (2, 16, 128, 128)

    tileshape = Shape(
        (16,) + tuple(default_raw.shape.sig),
        sig_dims=default_raw.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_raw.shape,
    )

    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)

    # ~1MB
    assert tuple(t.tile_slice.shape) == (16, 128, 128)
    for _ in tiles:
        pass


def test_scheme_too_large(default_raw):
    partitions = default_raw.get_partitions()
    p = next(partitions)
    depth = p.shape[0]

    # we make a tileshape that is too large for the partition here:
    tileshape = Shape(
        (depth + 1,) + tuple(default_raw.shape.sig),
        sig_dims=default_raw.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=default_raw.shape,
    )

    # tile shape is clamped to partition shape:
    tiles = p.get_tiles(tiling_scheme=tiling_scheme)
    t = next(tiles)
    assert tuple(t.tile_slice.shape) == tuple((depth,) + default_raw.shape.sig)
    for _ in tiles:
        pass


def test_comparison(default_raw, default_raw_data, lt_ctx_fast):
    udf = ValidationUDF(
        reference=reshaped_view(default_raw_data, (-1, *tuple(default_raw.shape.sig)))
    )
    lt_ctx_fast.run_udf(udf=udf, dataset=default_raw)


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_comparison_roi(default_raw, default_raw_data, lt_ctx_fast, as_sparse):
    roi = np.random.choice(
        [True, False],
        size=tuple(default_raw.shape.nav),
        p=[0.5, 0.5]
    )
    udf = ValidationUDF(reference=default_raw_data[roi])
    if as_sparse:
        roi = roi_as_sparse(roi)
    lt_ctx_fast.run_udf(udf=udf, dataset=default_raw, roi=roi)


def test_pickle_is_small(default_raw):
    pickled = cloudpickle.dumps(default_raw)
    cloudpickle.loads(pickled)

    # let's keep the pickled dataset size small-ish:
    assert len(pickled) < 2 * 1024


def test_apply_mask_analysis(default_raw, lt_ctx):
    mask = np.ones((128, 128))
    analysis = lt_ctx.create_mask_analysis(factories=[lambda: mask], dataset=default_raw)
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (16, 16)


def test_sum_analysis(default_raw, lt_ctx):
    analysis = lt_ctx.create_sum_analysis(dataset=default_raw)
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (128, 128)


def test_pick_analysis(default_raw, lt_ctx):
    analysis = PickFrameAnalysis(dataset=default_raw, parameters={"x": 15, "y": 15})
    results = lt_ctx.run(analysis)
    assert results[0].raw_data.shape == (128, 128)
    assert np.count_nonzero(results[0].raw_data) > 0


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_correction_default(default_raw, lt_ctx, with_roi, as_sparse):
    ds = default_raw

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None
    if with_roi and as_sparse:
        roi = roi_as_sparse(roi)
    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


@pytest.mark.with_numba
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_roi_1(default_raw, lt_ctx, as_sparse):
    p = next(default_raw.get_partitions())
    roi = np.zeros(p.meta.shape.flatten_nav().nav, dtype=bool)
    roi[0] = 1
    tiles = []
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=Shape((1, 128, 128), sig_dims=2),
        dataset_shape=default_raw.shape,
    )
    if as_sparse:
        roi = roi_as_sparse(roi)
    for tile in p.get_tiles(tiling_scheme=tiling_scheme, dest_dtype="float32", roi=roi):
        print("tile:", tile)
        tiles.append(tile)
    assert len(tiles) == 1
    assert tiles[0].tile_slice.origin == (0, 0, 0)
    assert tuple(tiles[0].tile_slice.shape) == (1, 128, 128)


@pytest.mark.with_numba
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_roi_2(default_raw, lt_ctx, as_sparse):
    p = next(default_raw.get_partitions())
    roi = np.zeros(p.meta.shape.flatten_nav(), dtype=bool)
    stackheight = 4
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=Shape((stackheight, 128, 128), sig_dims=2),
        dataset_shape=default_raw.shape,
    )
    roi[0:stackheight + 2] = 1
    if as_sparse:
        roi = roi_as_sparse(roi)
    tiles = p.get_tiles(tiling_scheme=tiling_scheme, dest_dtype="float32", roi=roi)
    tiles = list(tiles)


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_uint16_as_float32(uint16_raw, lt_ctx, as_sparse):
    p = next(uint16_raw.get_partitions())
    roi = np.zeros(p.meta.shape.flatten_nav(), dtype=bool)

    stackheight = 4
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=Shape((stackheight, 128, 128), sig_dims=2),
        dataset_shape=uint16_raw.shape,
    )
    roi[0:stackheight + 2] = 1
    if as_sparse:
        roi = roi_as_sparse(roi)
    tiles = p.get_tiles(tiling_scheme=tiling_scheme, dest_dtype="float32", roi=roi)
    tiles = list(tiles)


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_correction_uint16(uint16_raw, lt_ctx, with_roi, as_sparse):
    ds = uint16_raw

    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
    else:
        roi = None
    if with_roi and as_sparse:
        roi = roi_as_sparse(roi)
    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


def test_macrotile_normal(lt_ctx, default_raw):
    ps = default_raw.get_partitions()
    _ = next(ps)
    p2 = next(ps)
    macrotile = p2.get_macrotile()
    assert macrotile.tile_slice.shape == p2.shape
    assert macrotile.tile_slice.origin[0] == p2._start_frame


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_macrotile_roi_1(lt_ctx, default_raw, as_sparse):
    roi = np.zeros(default_raw.shape.nav, dtype=bool)
    roi[0, 5] = 1
    roi[0, 1] = 1
    p = next(default_raw.get_partitions())
    if as_sparse:
        roi = roi_as_sparse(roi)
    macrotile = p.get_macrotile(roi=roi)
    assert tuple(macrotile.tile_slice.shape) == (2, 128, 128)


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_macrotile_roi_2(lt_ctx, default_raw, as_sparse):
    roi = np.zeros(default_raw.shape.nav, dtype=bool)
    # all ones are in the first partition, so we don't get any data in p2:
    roi[0, 5] = 1
    roi[0, 1] = 1
    ps = default_raw.get_partitions()
    _ = next(ps)
    p2 = next(ps)

    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=p2.shape,
        dataset_shape=default_raw.shape,
    )
    if as_sparse:
        roi = roi_as_sparse(roi)
    p2._get_read_ranges(tiling_scheme, roi=None)
    p2._get_read_ranges(tiling_scheme, roi=roi)

    macrotile = p2.get_macrotile(roi=roi)
    assert tuple(macrotile.tile_slice.shape) == (0, 128, 128)


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_macrotile_roi_3(lt_ctx, default_raw, as_sparse):
    roi = np.ones(default_raw.shape.nav, dtype=bool)
    ps = default_raw.get_partitions()
    _ = next(ps)
    p2 = next(ps)
    if as_sparse:
        roi = roi_as_sparse(roi)
    macrotile = p2.get_macrotile(roi=roi)
    assert tuple(macrotile.tile_slice.shape) == tuple(p2.shape)


def test_cache_key_json_serializable(default_raw):
    json.dumps(default_raw.get_cache_key())


def test_message_converter_direct():
    src = {
        "type": "RAW",
        "path": "p",
        "dtype": "d",
        "nav_shape": [16, 16],
        "sig_shape": [8, 8],
        "sync_offset": 0,
    }
    converted = RAWDatasetParams().convert_to_python(src)
    assert converted == {
        "path": "p",
        "dtype": "d",
        "nav_shape": [16, 16],
        "sig_shape": [8, 8],
        "sync_offset": 0,
    }


@pytest.mark.dist
def test_raw_on_workers(raw_on_workers, dist_ctx):
    # should not exist on the client node:
    assert not os.path.exists(raw_on_workers._path)
    res = dist_ctx.executor.run_each_host(os.path.exists, raw_on_workers._path)
    assert len(res) == 2
    assert all(res)


@pytest.mark.dist
def test_sum_on_dist(raw_on_workers, dist_ctx):
    print(dist_ctx.executor.run_each_host(lambda: os.system("hostname")))
    print(dist_ctx.executor.get_available_workers().group_by_host())
    print(dist_ctx.executor.get_available_workers())
    print(dist_ctx.executor.run_each_host(
        lambda: os.listdir(os.path.dirname(raw_on_workers._path))))
    analysis = dist_ctx.create_sum_analysis(dataset=raw_on_workers)
    results = dist_ctx.run(analysis)
    assert results[0].raw_data.shape == (128, 128)


def test_ctx_load_old(lt_ctx, default_raw, recwarn):
    lt_ctx.load(
        "raw",
        path=default_raw._path,
        nav_shape=(16, 16),
        dtype="float32",
        detector_size_raw=(128, 128),
        crop_detector_to=(128, 128)
    )
    assert len(recwarn) == 2
    w0 = recwarn.pop(FutureWarning)
    assert issubclass(w0.category, FutureWarning)
    w1 = recwarn.pop(FutureWarning)
    assert issubclass(w1.category, FutureWarning)


def test_missing_sig_shape(lt_ctx, default_raw):
    with pytest.raises(TypeError) as e:
        lt_ctx.load(
            "raw",
            path=default_raw._path,
            nav_shape=(16, 16),
            dtype="float32",
        )
    assert e.match("missing 1 required argument: 'sig_shape'")


@pytest.mark.skipif(
    sys.platform.startswith("darwin"),
    reason="No support for direct I/O on Mac OS X"
)
def test_load_direct(lt_ctx, default_raw):
    ds_direct = lt_ctx.load(
        "raw",
        path=default_raw._path,
        nav_shape=(16, 16),
        sig_shape=(16, 16),
        dtype="float32",
        io_backend=DirectBackend(),
    )
    analysis = lt_ctx.create_sum_analysis(dataset=ds_direct)
    lt_ctx.run(analysis)


@pytest.mark.skipif(
    sys.platform.startswith("darwin"),
    reason="No support for direct I/O on Mac OS X"
)
def test_load_legacy_direct(lt_ctx, default_raw):
    with pytest.warns(FutureWarning):
        ds_direct = lt_ctx.load(
            "raw",
            path=default_raw._path,
            nav_shape=(16, 16),
            sig_shape=(16, 16),
            dtype="float32",
            enable_direct=True,
        )
    analysis = lt_ctx.create_sum_analysis(dataset=ds_direct)
    lt_ctx.run(analysis)


@pytest.mark.skipif(
    sys.platform.startswith("darwin"),
    reason="No support for direct I/O on Mac OS X"
)
def test_compare_direct_to_mmap(lt_ctx, default_raw, direct_raw):
    y = random.choice(range(default_raw.shape.nav[0]))
    x = random.choice(range(default_raw.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_raw,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=direct_raw,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


def test_big_endian(big_endian_raw, lt_ctx):
    udf = SumUDF()
    lt_ctx.run_udf(udf=udf, dataset=big_endian_raw)


@pytest.mark.parametrize(
    "with_roi", (True, False)
)
@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_correction_big_endian(big_endian_raw, lt_ctx, with_roi, as_sparse):
    ds = big_endian_raw
    if with_roi:
        roi = np.zeros(ds.shape.nav, dtype=bool)
        roi[:1] = True
        if as_sparse:
            roi = roi_as_sparse(roi)
    else:
        roi = None

    dataset_correction_verification(ds=ds, roi=roi, lt_ctx=lt_ctx)


@pytest.mark.parametrize(
    "io_backend", (
        BufferedBackend(),
        MMapBackend(),
    ),
)
def test_positive_sync_offset(lt_ctx, raw_dataset_8x8x8x8, raw_data_8x8x8x8_path, io_backend):
    udf = SumSigUDF()
    sync_offset = 2

    ds_with_offset = RawFileDataSet(
        path=raw_data_8x8x8x8_path,
        nav_shape=(8, 8),
        sig_shape=(8, 8),
        dtype="float32",
        sync_offset=sync_offset,
        io_backend=io_backend,
    )
    ds_with_offset.set_num_cores(4)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == 2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (4,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    tiles = p0.get_tiles(tiling_scheme)
    t0 = next(tiles)
    assert tuple(t0.tile_slice.origin) == (0, 0, 0)
    for _ in tiles:
        pass

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (48, 0, 0)
    assert p.slice.shape[0] == 16

    result = lt_ctx.run_udf(dataset=raw_dataset_8x8x8x8, udf=udf)
    result = result['intensity'].raw_data[sync_offset:]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[
        :ds_with_offset._meta.image_count - sync_offset
    ]

    assert np.allclose(result, result_with_offset)


@pytest.mark.parametrize(
    "io_backend", (
        BufferedBackend(),
        MMapBackend(),
    ),
)
def test_negative_sync_offset(lt_ctx, raw_dataset_8x8x8x8, raw_data_8x8x8x8_path, io_backend):
    udf = SumSigUDF()
    sync_offset = -2

    ds_with_offset = RawFileDataSet(
        path=raw_data_8x8x8x8_path,
        nav_shape=(8, 8),
        sig_shape=(8, 8),
        dtype="float32",
        sync_offset=sync_offset,
        io_backend=io_backend,
    )
    ds_with_offset.set_num_cores(4)
    ds_with_offset = ds_with_offset.initialize(lt_ctx.executor)
    ds_with_offset.check_valid()

    p0 = next(ds_with_offset.get_partitions())
    assert p0._start_frame == -2
    assert p0.slice.origin == (0, 0, 0)

    tileshape = Shape(
        (4,) + tuple(ds_with_offset.shape.sig),
        sig_dims=ds_with_offset.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds_with_offset.shape,
    )

    tiles = p0.get_tiles(tiling_scheme)
    t0 = next(tiles)
    assert tuple(t0.tile_slice.origin) == (2, 0, 0)
    for _ in tiles:
        pass

    for p in ds_with_offset.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p.slice.origin == (48, 0, 0)
    assert p.slice.shape[0] == 16

    result = lt_ctx.run_udf(dataset=raw_dataset_8x8x8x8, udf=udf)
    result = result['intensity'].raw_data[:raw_dataset_8x8x8x8._meta.image_count - abs(sync_offset)]

    result_with_offset = lt_ctx.run_udf(dataset=ds_with_offset, udf=udf)
    result_with_offset = result_with_offset['intensity'].raw_data[abs(sync_offset):]

    assert np.allclose(result, result_with_offset)


@pytest.mark.parametrize(
    "io_backend", (
        BufferedBackend(),
        MMapBackend(),
    ),
)
def test_missing_frames(lt_ctx, raw_data_8x8x8x8_path, io_backend):
    ds = RawFileDataSet(
        path=raw_data_8x8x8x8_path,
        nav_shape=(10, 8),
        sig_shape=(8, 8),
        dtype="float32",
        io_backend=io_backend,
    )
    ds.set_num_cores(4)
    ds = ds.initialize(lt_ctx.executor)

    tileshape = Shape(
        (4,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )

    for p in ds.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass

    assert p._start_frame == 60
    assert p._num_frames == 20
    assert p.slice.origin == (60, 0, 0)
    assert p.slice.shape[0] == 20
    assert t.tile_slice.origin == (60, 0, 0)
    assert t.tile_slice.shape[0] == 4


@pytest.mark.parametrize(
    "io_backend", (
        BufferedBackend(),
        MMapBackend(),
    ),
)
def test_too_many_frames(lt_ctx, raw_data_8x8x8x8_path, io_backend):
    ds = RawFileDataSet(
        path=raw_data_8x8x8x8_path,
        nav_shape=(6, 8),
        sig_shape=(8, 8),
        dtype="float32",
        io_backend=io_backend,
    )
    ds.set_num_cores(4)
    ds = ds.initialize(lt_ctx.executor)

    tileshape = Shape(
        (4,) + tuple(ds.shape.sig),
        sig_dims=ds.shape.sig.dims
    )
    tiling_scheme = TilingScheme.make_for_shape(
        tileshape=tileshape,
        dataset_shape=ds.shape,
    )

    for p in ds.get_partitions():
        for t in p.get_tiles(tiling_scheme=tiling_scheme):
            pass


def test_offset_smaller_than_image_count(lt_ctx, raw_data_8x8x8x8_path):
    sync_offset = -70

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "raw",
            path=raw_data_8x8x8x8_path,
            nav_shape=(8, 8),
            sig_shape=(8, 8),
            dtype="float32",
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-64, 64\), which is \(-image_count, image_count\)"
    )


def test_offset_greater_than_image_count(lt_ctx, raw_data_8x8x8x8_path):
    sync_offset = 70

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "raw",
            path=raw_data_8x8x8x8_path,
            nav_shape=(8, 8),
            sig_shape=(8, 8),
            dtype="float32",
            sync_offset=sync_offset
        )
    assert e.match(
        r"offset should be in \(-64, 64\), which is \(-image_count, image_count\)"
    )


@pytest.mark.parametrize(
    "io_backend", (
        BufferedBackend(),
        MMapBackend(),
    ),
)
def test_reshape_nav(lt_ctx, raw_dataset_8x8x8x8, raw_data_8x8x8x8_path, io_backend):
    udf = SumSigUDF()

    ds_with_1d_nav = lt_ctx.load(
        "raw",
        path=raw_data_8x8x8x8_path,
        nav_shape=(64,),
        sig_shape=(8, 8),
        dtype="float32",
        io_backend=io_backend,
    )
    result_with_1d_nav = lt_ctx.run_udf(dataset=ds_with_1d_nav, udf=udf)
    result_with_1d_nav = result_with_1d_nav['intensity'].raw_data

    result_with_2d_nav = lt_ctx.run_udf(dataset=raw_dataset_8x8x8x8, udf=udf)
    result_with_2d_nav = result_with_2d_nav['intensity'].raw_data

    ds_with_3d_nav = lt_ctx.load(
        "raw",
        path=raw_data_8x8x8x8_path,
        nav_shape=(2, 4, 8),
        sig_shape=(8, 8),
        dtype="float32",
    )
    result_with_3d_nav = lt_ctx.run_udf(dataset=ds_with_3d_nav, udf=udf)
    result_with_3d_nav = result_with_3d_nav['intensity'].raw_data

    assert np.allclose(result_with_1d_nav, result_with_2d_nav, result_with_3d_nav)


def test_too_large_sig_shape(lt_ctx, raw_data_8x8x8x8_path):
    nav_shape = (8, 8)
    sig_shape = (65, 65)

    with pytest.raises(Exception) as e:
        lt_ctx.load(
            "raw",
            path=raw_data_8x8x8x8_path,
            nav_shape=nav_shape,
            sig_shape=sig_shape,
            dtype="float32",
        )
    assert e.match(
        r"sig_shape must be less than size: 4096"
    )


def test_different_sig_shape(lt_ctx, raw_data_8x8x8x8_path):
    nav_shape = (8, 8)
    sig_shape = (4, 4)

    ds = lt_ctx.load(
        "raw",
        path=raw_data_8x8x8x8_path,
        nav_shape=nav_shape,
        sig_shape=sig_shape,
        dtype="float32",
    )

    assert ds._meta.image_count == 256


@pytest.mark.parametrize(
    "io_backend", (
        BufferedBackend(),
        MMapBackend(),
    ),
)
def test_extra_data_at_the_end(lt_ctx, raw_data_8x8x8x8_path, io_backend):
    """
    If there is extra data at the end of the file, make sure it is cut off
    """
    nav_shape = (8, 8)
    sig_shape = (3, 3)

    ds = lt_ctx.load(
        "raw",
        path=raw_data_8x8x8x8_path,
        nav_shape=nav_shape,
        sig_shape=sig_shape,
        dtype="float32",
        io_backend=io_backend,
    )

    assert ds._meta.image_count == 455


def test_scan_size_deprecation(lt_ctx, raw_data_8x8x8x8_path):
    scan_size = (2, 2)

    with pytest.warns(FutureWarning):
        ds = lt_ctx.load(
            "raw",
            path=raw_data_8x8x8x8_path,
            scan_size=scan_size,
            sig_shape=(8, 8),
            dtype="float32",
        )
    assert tuple(ds.shape) == (2, 2, 8, 8)


def test_detector_size_deprecation(lt_ctx, raw_data_8x8x8x8_path):
    detector_size = (8, 8)

    with pytest.warns(FutureWarning):
        ds = lt_ctx.load(
            "raw",
            path=raw_data_8x8x8x8_path,
            nav_shape=(8, 8),
            detector_size=detector_size,
            dtype="float32",
        )
    assert tuple(ds.shape) == (8, 8, 8, 8)


def test_compare_backends(lt_ctx, default_raw, buffered_raw):
    y = random.choice(range(default_raw.shape.nav[0]))
    x = random.choice(range(default_raw.shape.nav[1]))
    mm_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=default_raw,
        x=x, y=y,
    )).intensity
    buffered_f0 = lt_ctx.run(lt_ctx.create_pick_analysis(
        dataset=buffered_raw,
        x=x, y=y,
    )).intensity

    assert np.allclose(mm_f0, buffered_f0)


@pytest.mark.parametrize(
    "as_sparse", (
        False,
        True
    ),
)
def test_compare_backends_sparse(lt_ctx, default_raw, buffered_raw, as_sparse):
    roi = np.zeros(default_raw.shape.nav, dtype=bool).reshape((-1,))
    roi[0] = True
    roi[1] = True
    roi[-1] = True
    if as_sparse:
        roi = roi_as_sparse(roi)
    mm_f0 = lt_ctx.run_udf(dataset=default_raw, udf=PickUDF(), roi=roi)['intensity']
    buffered_f0 = lt_ctx.run_udf(dataset=buffered_raw, udf=PickUDF(), roi=roi)['intensity']

    assert np.allclose(mm_f0, buffered_f0)


# TODO: test for dataset with more than 2 sig dims

def test_diagnostics(default_raw):
    assert {"name": "dtype", "value": "float32"} in default_raw.get_diagnostics()


def test_bad_params(ds_params_tester, standard_bad_ds_params, raw_dataset_8x8x8x8):
    args = ("raw", raw_dataset_8x8x8x8._path)
    for params in standard_bad_ds_params:
        params['dtype'] = raw_dataset_8x8x8x8.meta.raw_dtype
        if 'nav_shape' not in params:
            params['nav_shape'] = raw_dataset_8x8x8x8.meta.shape.nav
        if 'sig_shape' not in params:
            params['sig_shape'] = raw_dataset_8x8x8x8.meta.shape.sig
        ds_params_tester(*args, **params)
